from contextlib import contextmanager

import pytest

from _pydev_bundle.pydev_override import overrides
from tests_python.debugger_fixtures import DebuggerRunnerSimple, debugger_runner_simple
from tests_python.debugger_unittest import AbstractWriterThread, SHOW_OTHER_DEBUG_INFO, \
    start_in_daemon_thread, wait_for_condition, IS_JYTHON
from _pydev_bundle.pydev_localhost import get_socket_names, get_socket_name
from _pydev_bundle.pydev_imports import xmlrpclib
from _pydev_bundle.pydev_imports import _queue as queue
from _pydev_bundle.pydev_imports import SimpleXMLRPCServer
import time
import socket

builtin_qualifier = "builtins"


@pytest.fixture
def console_setup(tmpdir):

    server_queue = queue.Queue()

    def notify_finished(more):
        server_queue.put(('notify_finished', more))
        return ''

    class ConsoleRunner(DebuggerRunnerSimple):

        @overrides(DebuggerRunnerSimple.add_command_line_args)
        def add_command_line_args(self, args):
            port, client_port = get_socket_names(2, close=True)
            args.extend((
                writer.get_pydevconsole_file(),
                str(port[1]),
                str(client_port[1])
            ))
            self.port = port
            self.client_port = client_port

            server = SimpleXMLRPCServer(client_port)
            server.register_function(notify_finished, "NotifyFinished")
            start_in_daemon_thread(server.serve_forever, [])

            self.proxy = xmlrpclib.ServerProxy("http://%s:%s/" % port)

            return args

    class WriterThread(AbstractWriterThread):

        if IS_JYTHON:
            EXPECTED_RETURNCODE = 'any'

        @overrides(AbstractWriterThread.additional_output_checks)
        def additional_output_checks(self, stdout, stderr):
            print('output found: %s - %s' % (stdout, stderr))

        @overrides(AbstractWriterThread.write_dump_threads)
        def write_dump_threads(self):
            pass  # no-op (may be called on timeout).

        def execute_line(self, command, more=False):
            runner.proxy.execLine(command)
            assert server_queue.get(timeout=5.) == ('notify_finished', more)

        def hello(self):

            def _hello():
                try:
                    msg = runner.proxy.hello('ignored')
                    if msg is not None:
                        if isinstance(msg, (list, tuple)):
                            msg = next(iter(msg))
                        if msg.lower().startswith('hello'):
                            return True
                except:
                    # That's ok, communication still not ready.
                    pass

                return False

            wait_for_condition(_hello)

        def close(self):
            try:
                runner.proxy.close()
            except:
                # Ignore any errors on close.
                pass

        def connect_to_debugger(self, debugger_port):
            runner.proxy.connectToDebugger(debugger_port)

    runner = ConsoleRunner(tmpdir)
    writer = WriterThread()

    class CaseSetup(object):

        @contextmanager
        def check_console(
                self,
                **kwargs
            ):
            for key, value in kwargs.items():
                assert hasattr(WriterThread, key)
                setattr(WriterThread, key, value)

            self.writer = writer

            args = runner.get_command_line()

            args = runner.add_command_line_args(args)

            if SHOW_OTHER_DEBUG_INFO:
                print('executing: %s' % (' '.join(args),))
            try:
                with runner.run_process(args, writer) as dct_with_stdout_stder:
                    writer.get_stdout = lambda: ''.join(dct_with_stdout_stder['stdout'])
                    writer.get_stderr = lambda: ''.join(dct_with_stdout_stder['stderr'])

                    # Make sure communication is setup.
                    writer.hello()
                    yield writer
            finally:
                writer.log = []

            stdout = dct_with_stdout_stder['stdout']
            stderr = dct_with_stdout_stder['stderr']
            writer.additional_output_checks(''.join(stdout), ''.join(stderr))

    return CaseSetup()


def test_console_simple(console_setup):
    with console_setup.check_console() as writer:
        writer.execute_line('a = 10')
        writer.execute_line('print("TEST SUCEEDED")')
        writer.close()
        writer.finished_ok = True


def test_console_debugger_connected(console_setup):

    class _DebuggerWriterThread(AbstractWriterThread):

        FORCE_KILL_PROCESS_WHEN_FINISHED_OK = True

        def __init__(self):
            AbstractWriterThread.__init__(self)
            socket_name = get_socket_name(close=True)
            server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            server_socket.bind(socket_name)
            server_socket.listen(1)
            self.port = socket_name[1]
            self.__server_socket = server_socket

        def run(self):
            print('waiting for second process')
            self.sock, addr = self.__server_socket.accept()
            print('accepted second process')

            from tests_python.debugger_unittest import ReaderThread
            self.reader_thread = ReaderThread(self.sock)
            self.reader_thread.start()

            self._sequence = -1
            # initial command is always the version
            self.write_version()
            self.log.append('start_socket')
            self.write_make_initial_run()
            time.sleep(1)

            seq = self.write_list_threads()
            msg = self.wait_for_list_threads(seq)
            assert msg.thread['name'] == 'MainThread'
            assert msg.thread['id'] == 'console_main'

            self.write_get_frame('console_main', '1')
            self.wait_for_vars([
                [
                    '<var name="a" type="int" qualifier="%s" value="int: 10"' % (builtin_qualifier,),
                    '<var name="a" type="int"  value="int',  # jython
                ],
            ])

            self.finished_ok = True

    with console_setup.check_console() as writer:
        writer.execute_line('a = 10')

        debugger_writer_thread = _DebuggerWriterThread()
        debugger_writer_thread.start()
        writer.connect_to_debugger(debugger_writer_thread.port)

        wait_for_condition(lambda: debugger_writer_thread.finished_ok)
        writer.execute_line('print("TEST SUCEEDED")')

        writer.close()
        writer.finished_ok = True




# coding: utf-8
from __future__ import unicode_literals
from _pydev_bundle._pydev_completer import (isidentifier, extract_token_and_qualifier,
    TokenAndQualifier)


def test_isidentifier():
    assert isidentifier('abc')
    assert not isidentifier('<')
    assert not isidentifier('')
    assert isidentifier('áéíóú')


def test_extract_token_and_qualifier():

    assert extract_token_and_qualifier('tok', 0, 0) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('tok', 0, 1) == TokenAndQualifier('', 't')
    assert extract_token_and_qualifier('tok', 0, 2) == TokenAndQualifier('', 'to')
    assert extract_token_and_qualifier('tok', 0, 3) == TokenAndQualifier('', 'tok')
    assert extract_token_and_qualifier('tok', 0, 4) == TokenAndQualifier('', 'tok')

    assert extract_token_and_qualifier('tok.qual', 0, 0) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('tok.qual', 0, 1) == TokenAndQualifier('', 't')
    assert extract_token_and_qualifier('tok.qual', 0, 2) == TokenAndQualifier('', 'to')
    assert extract_token_and_qualifier('tok.qual', 0, 3) == TokenAndQualifier('', 'tok')

    assert extract_token_and_qualifier('tok.qual', 0, 4) == TokenAndQualifier('tok', '')
    assert extract_token_and_qualifier('tok.qual', 0, 5) == TokenAndQualifier('tok', 'q')
    assert extract_token_and_qualifier('tok.qual', 0, 6) == TokenAndQualifier('tok', 'qu')
    assert extract_token_and_qualifier('tok.qual', 0, 7) == TokenAndQualifier('tok', 'qua')
    assert extract_token_and_qualifier('tok.qual', 0, 8) == TokenAndQualifier('tok', 'qual')

    # out of range (column)
    assert extract_token_and_qualifier('tok.qual.qual2', 0, 100) == TokenAndQualifier('tok.qual', 'qual2')

    assert extract_token_and_qualifier('t<ok', 0, 0) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('t<ok', 0, 1) == TokenAndQualifier('', 't')
    assert extract_token_and_qualifier('t<ok', 0, 2) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('t<ok', 0, 3) == TokenAndQualifier('', 'o')
    assert extract_token_and_qualifier('t<ok', 0, 4) == TokenAndQualifier('', 'ok')

    assert extract_token_and_qualifier('a\nt<ok', 1, 0) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('a\nt<ok', 1, 1) == TokenAndQualifier('', 't')
    assert extract_token_and_qualifier('a\nt<ok', 1, 2) == TokenAndQualifier('', '')
    assert extract_token_and_qualifier('a\nt<ok', 1, 3) == TokenAndQualifier('', 'o')
    assert extract_token_and_qualifier('a\nt<ok', 1, 4) == TokenAndQualifier('', 'ok')

    # out of range (line)
    assert extract_token_and_qualifier('a\nt<ok', 5, 4) == TokenAndQualifier('', '')



def full_function():
    # Note that this function is not called, it's there just to make the mapping explicit.
    a = 1  # map to cEll1, line 2
    b = 2  # map to cEll1, line 3

    c = 3  # map to cEll2, line 2
    d = 4  # map to cEll2, line 3


def create_code():
    cell1_code = compile(''' # line 1
a = 1  # line 2
b = 2  # line 3
''', '<cEll1>', 'exec')

    cell2_code = compile('''# line 1
c = 3  # line 2
d = 4  # line 3
''', '<cEll2>', 'exec')

    # Set up the source in linecache. Python doesn't have a public API for
    # this, so we have to hack around it, similar to what IPython does.
    import linecache
    import time
    code = ''' # line 1
a = 1  # line 2
b = 2  # line 3
'''
    linecache.cache['<cEll1>'] = (
        len(code),
        time.time(),
        [line + '\n' for line in code.splitlines()],
        '<cEll1>',
    )

    code = '''# line 1
c = 3  # line 2
d = 4  # line 3
'''
    linecache.cache['<cEll2>'] = (
        len(code),
        time.time(),
        [line + '\n' for line in code.splitlines()],
        '<cEll2>',
    )

    return {'cEll1': cell1_code, 'cEll2': cell2_code}


if __name__ == '__main__':
    code = create_code()
    exec(code['cEll1'])
    exec(code['cEll1'])

    exec(code['cEll2'])
    exec(code['cEll2'])
    print('TEST SUCEEDED')



from typing import Optional, Hashable, TypeVar, Generic, Iterator, Tuple, Set
from robotframework_ls.impl.protocols import (
    IRobotDocument,
    ICompletionContextWorkspaceCaches,
    ICompletionContextDependencyGraph,
    IOnDependencyChanged,
)
from robocorp_ls_core import uris
from collections import OrderedDict
import threading
from robotframework_ls.impl.robot_constants import (
    ROBOT_AND_TXT_FILE_EXTENSIONS,
    LIBRARY_FILE_EXTENSIONS,
)
from contextlib import contextmanager
from robocorp_ls_core.options import BaseOptions
from robocorp_ls_core.robotframework_log import get_logger
import json

T = TypeVar("T")

log = get_logger(__name__)


class _LRU(Generic[T]):
    """
    A really-simple LRU to help us keep track of only a few dependency graphs.
    """

    def __init__(self, max_size: int):
        self._cache: "OrderedDict[Hashable, T]" = OrderedDict()
        self._max_size = max_size

    def get(self, key: Hashable) -> Optional[T]:
        val = self._cache.get(key)
        if val is None:
            return None
        else:
            self._cache.move_to_end(key)
            return val

    def put(self, key: Hashable, value: T) -> None:
        self._cache[key] = value
        self._cache.move_to_end(key)
        if len(self._cache) > self._max_size:
            self._cache.popitem(last=False)

    def pop(self, key: Hashable, default: Optional[T] = None) -> Optional[T]:
        return self._cache.pop(key, default)

    def clear(self) -> None:
        self._cache.clear()

    def items(self) -> Iterator[Tuple[Hashable, T]]:
        yield from self._cache.items()

    def values(self) -> Iterator[T]:
        yield from self._cache.values()


class _InvalidationTracker:
    def __init__(self):
        self._uris_invalidated = set()
        self._all_invalidated = False

    def mark_uri_invalidated(self, uri):
        self._uris_invalidated.add(uri)

    def mark_all_invalidated(self):
        self._all_invalidated = True

    def is_dependency_graph_still_valid(self, dependency_graph) -> bool:
        if self._all_invalidated:
            return False
        for uri in self._uris_invalidated:
            if dependency_graph.do_invalidate_on_uri_change(uri):
                return False

        return True


class CompletionContextWorkspaceCaches:
    def __init__(
        self, on_dependency_changed: Optional[IOnDependencyChanged] = None
    ) -> None:
        self._lock = threading.Lock()
        # Small cache because invalidation could become slow in a big cache
        # (and it should be enough to hold what we're currently working with).
        self._cached: _LRU[ICompletionContextDependencyGraph] = _LRU(5)
        self.cache_hits = 0
        self.invalidations = 0

        self._invalidation_trackers: Set[_InvalidationTracker] = set()
        self._on_dependency_changed = on_dependency_changed

    def _invalidate_uri(self, uri: str) -> None:
        with self._lock:
            notified = set()
            for invalidation_tracker in self._invalidation_trackers:
                invalidation_tracker.mark_uri_invalidated(uri)

            did_invalidate_entry = False

            for key, entry in tuple(self._cached.items()):
                if entry.do_invalidate_on_uri_change(uri):

                    uri = entry.get_root_doc().uri
                    if uri not in notified:
                        notified.add(uri)
                        if self._on_dependency_changed:
                            self._on_dependency_changed(uri)

                    did_invalidate_entry = True
                    invalidated: Optional[
                        ICompletionContextDependencyGraph
                    ] = self._cached.pop(key, None)
                    if BaseOptions.DEBUG_CACHE_DEPS and invalidated:
                        log.info(
                            "Invalidated: %s\n%s\n",
                            key,
                            json.dumps(invalidated.to_dict(), indent=4),
                        )

            if BaseOptions.DEBUG_CACHE_DEPS and not did_invalidate_entry:
                log.info("%s did not invalidate the caches:", uri)
                for key, entry in tuple(self._cached.items()):
                    log.info(
                        json.dumps(entry.to_dict(), indent=4),
                    )

    @contextmanager
    def invalidation_tracker(self):
        try:
            with self._lock:
                invalidation_tracker = _InvalidationTracker()
                self._invalidation_trackers.add(invalidation_tracker)

            yield invalidation_tracker

        finally:
            with self._lock:
                self._invalidation_trackers.discard(invalidation_tracker)

    def on_file_changed(self, filename: str):
        """
        Called when a file is changed in the file-system (i.e.: it was saved).
        """
        if filename:
            lower = filename.lower()
            if lower.endswith(ROBOT_AND_TXT_FILE_EXTENSIONS):
                uri = uris.from_fs_path(filename)
                self._invalidate_uri(uri)

            elif lower.endswith(LIBRARY_FILE_EXTENSIONS):
                # If a library changes, we consider all caches invalid because
                # we don't hold an association to know which library maps to
                # which files at this level.
                self.clear_caches()

    def on_updated_document(self, uri: str, document: Optional[IRobotDocument]):
        """
        Called when a (robot) document was updated in-memory.
        :param uri:
            The uri for the document that was just updated.
        :param document:
            The document just updated or None if it was removed.
        """
        self._invalidate_uri(uri)

    def clear_caches(self):
        """
        Called when all caches should be cleared.
        """
        with self._lock:
            self.invalidations += 1
            for invalidation_tracker in self._invalidation_trackers:
                invalidation_tracker.mark_all_invalidated()
            self._cached.clear()

    def dispose(self):
        self.clear_caches()

    def get_cached_dependency_graph(
        self, cache_key: Hashable
    ) -> Optional[ICompletionContextDependencyGraph]:
        with self._lock:
            ret = self._cached.get(cache_key)
            if ret is not None:
                self.cache_hits += 1

            if BaseOptions.DEBUG_CACHE_DEPS:
                if ret is not None:

                    log.info(
                        "Cache HIT (%s):\n%s\n",
                        cache_key,
                        json.dumps(ret.to_dict(), indent=4),
                    )
                else:
                    log.info(
                        "Cache MISS (%s):\n%s\n",
                        cache_key,
                        "  \n".join(
                            json.dumps(x.to_dict(), indent=4)
                            for x in self._cached.values()
                        ),
                    )

            return ret

    def cache_dependency_graph(
        self,
        cache_key: Hashable,
        dependency_graph: ICompletionContextDependencyGraph,
        invalidation_tracker: _InvalidationTracker,
    ) -> None:
        with self._lock:
            if invalidation_tracker.is_dependency_graph_still_valid(dependency_graph):
                self._cached.put(cache_key, dependency_graph)

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ICompletionContextWorkspaceCaches = check_implements(self)



from typing import Iterator, Tuple, Optional, Deque, Dict, Sequence, List, Set

from robocorp_ls_core.ordered_set import OrderedSet
from robotframework_ls.impl.protocols import (
    ICompletionContextDependencyGraph,
    LibraryDependencyInfo,
    ICompletionContext,
    IResourceImportNode,
    IRobotDocument,
    ICompletionContextWorkspaceCaches,
    IVariableImportNode,
)
from robotframework_ls.impl.robot_constants import BUILTIN_LIB
from robocorp_ls_core.callbacks import Callback
from robocorp_ls_core import uris


class _Memo(object):
    def __init__(self):
        self.clear()

    def clear(self):
        self._followed_imports_variables = {}
        self._followed_imports = {}
        self._completed_libraries = {}

    def follow_import(self, uri: str) -> bool:
        if uri not in self._followed_imports:
            self._followed_imports[uri] = True
            return True

        return False

    def follow_import_variables(self, uri: str) -> bool:
        if uri not in self._followed_imports_variables:
            self._followed_imports_variables[uri] = True
            return True

        return False

    def complete_for_library(self, library_name: str, alias: Optional[str]) -> bool:
        key = (library_name, alias)
        if key not in self._completed_libraries:
            self._completed_libraries[key] = True
            return True

        return False


class CompletionContextDependencyGraph:
    """
    This class is used to map dependencies from a given document

    i.e.: Imports / Resources / Variables
    """

    # Called as: on_before_cache_dependency_graph(ICompletionContextDependencyGraph)
    on_before_cache_dependency_graph = Callback()

    def __init__(self, root_doc: IRobotDocument):
        self._root_doc: IRobotDocument = root_doc
        self._doc_uri_to_library_infos: Dict[
            str, OrderedSet[LibraryDependencyInfo]
        ] = {}
        self._doc_uri_to_resource_imports: Dict[
            str, Sequence[Tuple[IResourceImportNode, Optional[IRobotDocument]]]
        ] = {}
        self._doc_uri_to_variable_imports: Dict[
            str, Sequence[Tuple[IVariableImportNode, Optional[IRobotDocument]]]
        ] = {}

        self._invalidate_on_uri_changes: Set[str] = set()

    def to_dict(self):
        libraries = {}
        for doc_uri, library_infos in self._doc_uri_to_library_infos.items():
            libraries[doc_uri] = [x.to_dict() for x in library_infos]

        resources = {}
        for doc_uri, resource_imports in self._doc_uri_to_resource_imports.items():
            resources[doc_uri] = [
                (doc.uri if doc else None) for (_node, doc) in resource_imports
            ]

        variables = {}
        for doc_uri, variable_imports in self._doc_uri_to_variable_imports.items():
            variables[doc_uri] = [
                (doc.uri if doc else None) for (_node, doc) in variable_imports
            ]

        ret = {
            "root_doc": self._root_doc.uri,
            "libraries": libraries,
            "resources": resources,
            "variables": variables,
        }
        return ret

    def print_repr(self):
        import json

        print(json.dumps(self.to_dict(), indent=4))

    def add_library_infos(
        self,
        doc_uri: str,
        library_infos: OrderedSet[LibraryDependencyInfo],
    ):
        self._doc_uri_to_library_infos[doc_uri] = library_infos
        self._invalidate_on_uri_changes.add(uris.normalize_uri(doc_uri))

    def add_resource_infos(
        self,
        doc_uri: str,
        resource_imports_as_docs: Sequence[
            Tuple[IResourceImportNode, Optional[IRobotDocument]]
        ],
    ):
        self._doc_uri_to_resource_imports[doc_uri] = resource_imports_as_docs

        self._invalidate_on_uri_changes.add(uris.normalize_uri(doc_uri))
        for _, resource_doc in resource_imports_as_docs:
            if resource_doc is not None:
                self._invalidate_on_uri_changes.add(
                    uris.normalize_uri(resource_doc.uri)
                )

    def add_variable_infos(
        self,
        doc_uri: str,
        new_variable_imports: Sequence[
            Tuple[IVariableImportNode, Optional[IRobotDocument]]
        ],
    ):
        self._doc_uri_to_variable_imports[doc_uri] = new_variable_imports

        self._invalidate_on_uri_changes.add(uris.normalize_uri(doc_uri))
        for _, variable_doc in new_variable_imports:
            if variable_doc is not None:
                self._invalidate_on_uri_changes.add(
                    uris.normalize_uri(variable_doc.uri)
                )

    def get_root_doc(self) -> IRobotDocument:
        return self._root_doc

    def iter_libraries(self, doc_uri: str) -> Iterator[LibraryDependencyInfo]:
        infos = self._doc_uri_to_library_infos.get(doc_uri)
        if infos:
            yield from infos

    def iter_all_libraries(self) -> Iterator[LibraryDependencyInfo]:
        for infos in self._doc_uri_to_library_infos.values():
            yield from infos

    def iter_resource_imports_with_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        infos = self._doc_uri_to_resource_imports.get(doc_uri)
        if infos:
            yield from infos

    def iter_all_resource_imports_with_docs(
        self,
    ) -> Iterator[Tuple[IResourceImportNode, Optional[IRobotDocument]]]:
        for infos in self._doc_uri_to_resource_imports.values():
            yield from infos

    def iter_variable_imports_as_docs(
        self, doc_uri: str
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        variable_imports = self._doc_uri_to_variable_imports.get(doc_uri)
        if variable_imports:
            yield from variable_imports

    def iter_all_variable_imports_as_docs(
        self,
    ) -> Iterator[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:
        for variable_imports in self._doc_uri_to_variable_imports.values():
            yield from variable_imports

    def do_invalidate_on_uri_change(self, uri: str) -> bool:
        if uris.normalize_uri(uri) == uris.normalize_uri(self.get_root_doc().uri):
            # Changes in the root don't invalidate the dependency info (rather
            # it's used in the cache key when checking so it won't be a match).
            return False

        return uris.normalize_uri(uri) in self._invalidate_on_uri_changes

    @classmethod
    def _collect_library_info_from_completion_context(
        cls, curr_ctx: ICompletionContext, is_root_context: bool, memo: _Memo
    ) -> OrderedSet[LibraryDependencyInfo]:
        from robotframework_ls.impl import ast_utils
        from robot.api import Token

        # Collect libraries information
        libraries = curr_ctx.get_imported_libraries()

        new_library_infos: OrderedSet[LibraryDependencyInfo] = OrderedSet()
        if is_root_context:
            new_library_infos.add(
                LibraryDependencyInfo(BUILTIN_LIB, None, True, None, None)
            )

        for library in libraries:
            name_tok = library.get_token(Token.NAME)

            if name_tok and name_tok.value:
                alias = library.alias
                args = ast_utils.get_library_arguments_serialized(library)
                node = library

                lib_info = LibraryDependencyInfo(
                    curr_ctx.token_value_resolving_variables(name_tok),
                    alias,
                    False,
                    args,
                    node,
                )
                if not memo.complete_for_library(lib_info.name, lib_info.alias):
                    continue

                new_library_infos.add(lib_info)
        return new_library_infos

    @classmethod
    def _collect_resource_info_from_completion_context(
        cls, curr_ctx: ICompletionContext, is_root_context: bool, memo: _Memo
    ):
        # Collect related resource imports information.
        resource_imports_as_docs = curr_ctx.get_resource_imports_as_docs()
        new_resource_infos: List[
            Tuple[IResourceImportNode, Optional[IRobotDocument]]
        ] = []
        if resource_imports_as_docs:
            for resource_import_node, resource_doc in resource_imports_as_docs:
                if resource_doc is None:
                    if is_root_context:
                        # We need to keep the empty nodes for the initial context.
                        new_resource_infos.append((resource_import_node, resource_doc))
                elif memo.follow_import(resource_doc.uri):
                    new_resource_infos.append((resource_import_node, resource_doc))
        return new_resource_infos

    @classmethod
    def _collect_variable_info_from_completion_context(
        cls, curr_ctx: ICompletionContext, is_root_context: bool, memo: _Memo
    ) -> Sequence[Tuple[IVariableImportNode, Optional[IRobotDocument]]]:

        variable_imports_as_docs = curr_ctx.get_variable_imports_as_docs()
        new_variable_infos: List[
            Tuple[IVariableImportNode, Optional[IRobotDocument]]
        ] = []
        if variable_imports_as_docs:
            for variable_import_node, variable_doc in variable_imports_as_docs:
                if variable_doc is None:
                    if is_root_context:
                        # We need to keep the empty nodes for the initial context.
                        new_variable_infos.append((variable_import_node, variable_doc))
                elif memo.follow_import(variable_doc.uri):
                    new_variable_infos.append((variable_import_node, variable_doc))
        return new_variable_infos

    @classmethod
    def from_completion_context(
        cls, completion_context: ICompletionContext
    ) -> ICompletionContextDependencyGraph:
        from collections import deque
        from robot.api import Token

        caches: ICompletionContextWorkspaceCaches = (
            completion_context.workspace.completion_context_workspace_caches
        )

        memo = _Memo()
        with caches.invalidation_tracker() as invalidation_tracker:
            dependency_graph = CompletionContextDependencyGraph(completion_context.doc)

            initial_library_infos = cls._collect_library_info_from_completion_context(
                completion_context, is_root_context=True, memo=memo
            )

            # i.e.: Note that the cache key involves the import names, not the
            # import locations (so, in case there's a match we need to fix
            # the newly returned info).
            cache_key = (
                completion_context.doc.uri,
                tuple(
                    (info.name, info.alias, info.builtin, info.args)
                    for info in initial_library_infos
                ),
                tuple(
                    (
                        tuple(t.value for t in node.tokens if t.type == t.NAME)
                        for node in completion_context.get_resource_imports()
                    )
                ),
                tuple(
                    (
                        tuple(t.value for t in node.tokens if t.type == t.NAME)
                        for node in completion_context.get_variable_imports()
                    )
                ),
            )

            found = caches.get_cached_dependency_graph(cache_key)
            if found is not None:
                # We need to fix the nodes as we can match even when the node
                # lines/columns change.

                # Library infos for the root can be added again as is because we
                # just collected the whole info again (including the nodes).
                found.add_library_infos(
                    completion_context.doc.uri, initial_library_infos
                )

                # For resources, we don't want to add them completely because this
                # means that the resolution must be done again.
                names_to_resources: Dict[
                    Tuple[str, ...], List[IResourceImportNode]
                ] = {}

                for resource_import in completion_context.get_resource_imports():
                    key = tuple(x.value for x in resource_import.get_tokens(Token.NAME))
                    names_to_resources.setdefault(key, []).append(resource_import)

                new_resources: List[
                    Tuple[IResourceImportNode, Optional[IRobotDocument]]
                ] = []
                for (
                    resource_import,
                    resource_doc,
                ) in found.iter_resource_imports_with_docs(completion_context.doc.uri):
                    key = tuple(x.value for x in resource_import.get_tokens(Token.NAME))
                    lst = names_to_resources.get(key)
                    if lst:
                        resource_import = lst.pop()
                        new_resources.append((resource_import, resource_doc))

                found.add_resource_infos(completion_context.doc.uri, new_resources)

                # Variables don't need any change as we don't store the nodes.
                return found

            completion_context_stack: Deque[ICompletionContext] = deque()
            completion_context_stack.append(completion_context)

            is_root_context = True

            # Mark as being followed.
            memo.follow_import(completion_context.doc.uri)

            while completion_context_stack:
                curr_ctx = completion_context_stack.popleft()

                if is_root_context:
                    # use what's been already collected to compute the cache.
                    new_library_infos = initial_library_infos
                else:
                    new_library_infos = (
                        cls._collect_library_info_from_completion_context(
                            curr_ctx, is_root_context, memo
                        )
                    )
                new_resource_infos = cls._collect_resource_info_from_completion_context(
                    curr_ctx, is_root_context, memo
                )
                new_variable_imports = (
                    cls._collect_variable_info_from_completion_context(
                        curr_ctx, is_root_context, memo
                    )
                )

                if new_library_infos:
                    dependency_graph.add_library_infos(
                        curr_ctx.doc.uri, new_library_infos
                    )

                if new_resource_infos:
                    dependency_graph.add_resource_infos(
                        curr_ctx.doc.uri, new_resource_infos
                    )
                    for _resource_import_node, resource_doc in new_resource_infos:
                        if resource_doc is not None:
                            new_ctx = curr_ctx.create_copy(resource_doc)
                            completion_context_stack.append(new_ctx)

                if new_variable_imports:
                    dependency_graph.add_variable_infos(
                        curr_ctx.doc.uri, new_variable_imports
                    )

                is_root_context = False

            cls.on_before_cache_dependency_graph(dependency_graph)
            caches.cache_dependency_graph(
                cache_key, dependency_graph, invalidation_tracker
            )
        return dependency_graph

    def __typecheckself__(self) -> None:
        from robocorp_ls_core.protocols import check_implements

        _: ICompletionContextDependencyGraph = check_implements(self)



import sys

IS_WIN = IS_WINDOWS = sys.platform == "win32"
IS_LINUX = sys.platform in ("linux", "linux2")
IS_MAC = sys.platform == "darwin"

IS_JYTHON = sys.platform.startswith("java")
IS_PY37_ONWARDS = sys.version_info[:2] >= (3, 7)


class Null:
    """
    Gotten from: http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/68205
    """

    def __init__(self, *args, **kwargs):
        return None

    def __call__(self, *args, **kwargs):
        return self

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        return self

    def __getattr__(self, mname):
        if len(mname) > 4 and mname[:2] == "__" and mname[-2:] == "__":
            # Don't pretend to implement special method names.
            raise AttributeError(mname)
        return self

    def __setattr__(self, name, value):
        return self

    def __delattr__(self, name):
        return self

    def __repr__(self):
        return "<Null>"

    def __str__(self):
        return "Null"

    def __len__(self):
        return 0

    def __getitem__(self, *args, **kwargs):
        return self

    def __setitem__(self, *args, **kwargs):
        pass

    def write(self, *args, **kwargs):
        pass

    def __nonzero__(self):
        return 0

    def __iter__(self):
        return iter(())


# Default instance
NULL = Null()



import unittest

class StillYetAnotherSampleTest(unittest.TestCase):
    
    def setUp(self):
        return

    def tearDown(self):
        return

    def test_non_unique_name(self):
        pass


if __name__ == '__main__':
    unittest.main()



import time


class ProceedContainer:
    proceed = False


def exit_while_loop():
    ProceedContainer.proceed = True
    return 'ok'


def sleep():
    while not ProceedContainer.proceed:  # The debugger should change the proceed to True to exit the loop.
        time.sleep(.1)


if __name__ == '__main__':
    sleep()

    print('TEST SUCEEDED')




from _pydevd_bundle.pydevd_extension_api import TypeResolveProvider
from _pydevd_bundle.pydevd_resolver import defaultResolver, MAX_ITEMS_TO_HANDLE, TOO_LARGE_ATTR, TOO_LARGE_MSG
from .pydevd_helpers import find_mod_attr


class NdArrayItemsContainer(object):
    pass


class NDArrayTypeResolveProvider(object):
    '''
    This resolves a numpy ndarray returning some metadata about the NDArray
    '''

    def can_provide(self, type_object, type_name):
        nd_array = find_mod_attr('numpy', 'ndarray')
        return nd_array is not None and issubclass(type_object, nd_array)

    def is_numeric(self, obj):
        if not hasattr(obj, 'dtype'):
            return False
        return obj.dtype.kind in 'biufc'

    def resolve(self, obj, attribute):
        if attribute == '__internals__':
            return defaultResolver.get_dictionary(obj)
        if attribute == 'min':
            if self.is_numeric(obj) and obj.size > 0:
                return obj.min()
            else:
                return None
        if attribute == 'max':
            if self.is_numeric(obj) and obj.size > 0:
                return obj.max()
            else:
                return None
        if attribute == 'shape':
            return obj.shape
        if attribute == 'dtype':
            return obj.dtype
        if attribute == 'size':
            return obj.size
        if attribute.startswith('['):
            container = NdArrayItemsContainer()
            i = 0
            format_str = '%0' + str(int(len(str(len(obj))))) + 'd'
            for item in obj:
                setattr(container, format_str % i, item)
                i += 1
                if i > MAX_ITEMS_TO_HANDLE:
                    setattr(container, TOO_LARGE_ATTR, TOO_LARGE_MSG)
                    break
            return container
        return None

    def get_dictionary(self, obj):
        ret = dict()
        ret['__internals__'] = defaultResolver.get_dictionary(obj)
        if obj.size > 1024 * 1024:
            ret['min'] = 'ndarray too big, calculating min would slow down debugging'
            ret['max'] = 'ndarray too big, calculating max would slow down debugging'
        elif obj.size == 0:
            ret['min'] = 'array is empty'
            ret['max'] = 'array is empty'
        else:
            if self.is_numeric(obj):
                ret['min'] = obj.min()
                ret['max'] = obj.max()
            else:
                ret['min'] = 'not a numeric object'
                ret['max'] = 'not a numeric object'
        ret['shape'] = obj.shape
        ret['dtype'] = obj.dtype
        ret['size'] = obj.size
        try:
            ret['[0:%s] ' % (len(obj))] = list(obj[0:MAX_ITEMS_TO_HANDLE])
        except:
            # This may not work depending on the array shape.
            pass
        return ret


import sys

if not sys.platform.startswith("java"):
    TypeResolveProvider.register(NDArrayTypeResolveProvider)



from robocorp_ls_core.python_ls import PythonLanguageServer
from robocorp_ls_core.basic import overrides
from robocorp_ls_core.robotframework_log import get_logger
from typing import Optional, List, Dict, Deque, Tuple, Sequence
from robocorp_ls_core.protocols import IConfig, IMonitor, ITestInfoTypedDict, IWorkspace
from functools import partial
from robocorp_ls_core.jsonrpc.endpoint import require_monitor
from robocorp_ls_core.lsp import (
    SymbolInformationTypedDict,
    FoldingRangeTypedDict,
    HoverTypedDict,
    TextDocumentTypedDict,
    CodeLensTypedDict,
    DocumentSymbolTypedDict,
    PositionTypedDict,
    LocationTypedDict,
    DocumentHighlightTypedDict,
    RangeTypedDict,
    CompletionItemTypedDict,
)
from robotframework_ls.impl.protocols import (
    IKeywordFound,
    IDefinition,
    ICompletionContext,
    EvaluatableExpressionTypedDict,
    IVariablesFromArgumentsFileLoader,
)
from robocorp_ls_core.watchdog_wrapper import IFSObserver
import itertools
import typing
import sys
import threading
import weakref


log = get_logger(__name__)


def complete_all(
    completion_context: ICompletionContext,
) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl import section_name_completions
    from robotframework_ls.impl import keyword_completions
    from robotframework_ls.impl import variable_completions
    from robotframework_ls.impl import dictionary_completions
    from robotframework_ls.impl import filesystem_section_completions
    from robotframework_ls.impl import keyword_parameter_completions
    from robotframework_ls.impl import auto_import_completions
    from robotframework_ls.impl.collect_keywords import (
        collect_keyword_name_to_keyword_found,
    )
    from robotframework_ls.impl import ast_utils

    ret = section_name_completions.complete(completion_context)
    if not ret:
        ret.extend(filesystem_section_completions.complete(completion_context))

    if not ret:
        token_info = completion_context.get_current_token()
        if token_info is not None:
            token = ast_utils.get_keyword_name_token(
                token_info.stack, token_info.node, token_info.token
            )
            if token is not None:
                keyword_name_to_keyword_found: Dict[
                    str, List[IKeywordFound]
                ] = collect_keyword_name_to_keyword_found(completion_context)
                ret.extend(keyword_completions.complete(completion_context))
                ret.extend(
                    auto_import_completions.complete(
                        completion_context, keyword_name_to_keyword_found
                    )
                )
                return ret

    if not ret:
        ret.extend(variable_completions.complete(completion_context))

    if not ret:
        ret.extend(dictionary_completions.complete(completion_context))

    if not ret:
        ret.extend(keyword_parameter_completions.complete(completion_context))

    return ret


class RobotFrameworkServerApi(PythonLanguageServer):
    """
    This is a custom server. It uses the same message-format used in the language
    server but with custom messages (i.e.: this is not the language server, but
    an API to use the bits we need from robotframework in a separate process).
    """

    def __init__(
        self,
        read_from,
        write_to,
        libspec_manager=None,
        observer: Optional[IFSObserver] = None,
        pre_generate_libspecs: bool = False,
        index_workspace: bool = False,
        collect_tests: bool = False,
    ):
        from robotframework_ls.impl.libspec_manager import LibspecManager

        PythonLanguageServer.__init__(self, read_from, write_to)
        self._index_workspace = index_workspace
        self._collect_tests = collect_tests

        if libspec_manager is None:
            try:
                libspec_manager = LibspecManager(
                    observer=observer,
                    endpoint=self._endpoint,
                    pre_generate_libspecs=pre_generate_libspecs,
                )
            except:
                log.exception("Unable to properly initialize the LibspecManager.")
                raise

        self.libspec_manager = libspec_manager
        self._version: Optional[str] = None
        self._logged_version_error = False
        self._next_time = partial(next, itertools.count(0))
        from collections import deque

        self._completion_contexts_saved_lock = threading.Lock()
        self._completion_contexts_saved: Deque[ICompletionContext] = deque()
        self._variables_from_arguments_files_loader: Sequence[
            IVariablesFromArgumentsFileLoader
        ] = []

    @overrides(PythonLanguageServer._create_config)
    def _create_config(self) -> IConfig:
        from robotframework_ls.robot_config import RobotConfig

        return RobotConfig()

    def m_version(self) -> str:
        """
        Kind of a code-smell:

        Either a string with the version is provided or an error message.

        It's now part of the API, so, keep as is.
        """
        if self._version is not None:
            return self._version
        try:
            import robot  # noqa
        except:
            version = (
                'Error in "import robot".\n'
                f"It seems that Robot Framework is not installed in {sys.executable}.\n"
                "Please install it in your environment and restart the Robot Framework Language Server\n"
                'or set: "robot.language-server.python" or "robot.python.executable"\n'
                "to point to a python installation that has Robot Framework installed.\n"
                "Hint: with pip it can be installed with:\n"
                f"{sys.executable} -m pip install robotframework\n"
            )
            log.exception(version)
        else:
            try:
                from robot import get_version

                version = get_version(naked=True)
            except:
                version = (
                    'Error calling "robot.get_version()".\n'
                    f"If the module: {robot}\n"
                    "is a module from your project, please rename it (as it is shadowing the Robot Framework `robot` package)\n"
                    "and restart the Robot Framework Language Server."
                )
                log.exception(version)
        self._version = version
        return self._version

    def _compute_min_version_error(self, min_version: Tuple[int, int]) -> Optional[str]:
        from robocorp_ls_core.basic import check_min_version

        v_or_error = self.m_version()
        try:
            # Check if it's an error or a valid version.
            tuple(int(x) for x in v_or_error.split("."))
        except:
            return v_or_error  # This is an error message

        if not check_min_version(v_or_error, min_version):
            return f"Expected Robot Framework version: {'.'.join((str(x) for x in min_version))}. Found: {v_or_error}"
        return None

    def _check_and_log_rf_dependency_version(self):
        error = self._compute_min_version_error((3, 2))
        if error is not None:
            if self._logged_version_error:
                log.info(
                    "Problem with `robotframework` dependency. See previous errors."
                )
            else:
                log.info(error)
                self._logged_version_error = True
            return False
        return True

    @overrides(PythonLanguageServer.m_workspace__did_change_configuration)
    def m_workspace__did_change_configuration(self, **kwargs):
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_VARIABLES_LOAD_FROM_ARGUMENTS_FILE,
        )

        PythonLanguageServer.m_workspace__did_change_configuration(self, **kwargs)
        self.libspec_manager.config = self.config

        try:
            variables_from_arguments_files = self.config.get_setting(
                OPTION_ROBOT_VARIABLES_LOAD_FROM_ARGUMENTS_FILE, str, ""
            ).strip()
            if not variables_from_arguments_files:
                self._variables_from_arguments_files_loader = []

            else:
                from robotframework_ls.impl.variables_from_arguments_file import (
                    VariablesFromArgumentsFileLoader,
                )

                created_variables_from_arguments_files = []

                for v in variables_from_arguments_files.split(","):
                    v = v.strip()
                    created_variables_from_arguments_files.append(
                        VariablesFromArgumentsFileLoader(v)
                    )
                self._variables_from_arguments_files_loader = (
                    created_variables_from_arguments_files
                )
        except:
            log.exception(
                f"Error getting options: {OPTION_ROBOT_VARIABLES_LOAD_FROM_ARGUMENTS_FILE}"
            )

    @overrides(PythonLanguageServer.lint)
    def lint(self, *args, **kwargs):
        pass  # No-op for this server.

    @overrides(PythonLanguageServer.cancel_lint)
    def cancel_lint(self, *args, **kwargs):
        pass  # No-op for this server.

    @overrides(PythonLanguageServer._obtain_fs_observer)
    def _obtain_fs_observer(self) -> IFSObserver:
        return self.libspec_manager.fs_observer

    @overrides(PythonLanguageServer._create_workspace)
    def _create_workspace(
        self, root_uri: str, fs_observer: IFSObserver, workspace_folders
    ) -> IWorkspace:
        from robotframework_ls.impl.robot_workspace import RobotWorkspace

        # Note: note done because our caches are removed promptly
        # for this to work it should be invalidate but the info
        # should be kept around so that we lint dependencies always
        # not just on the first cache invalidation.
        # WIP: test_dependency_graph_integration_lint
        # weak_self = weakref.ref(self)
        #
        # def on_dependency_changed(uri):
        #     s = weak_self()
        #     if s is not None:
        #         endpoint = s._endpoint
        #         if endpoint:
        #             endpoint.notify(
        #                 "$/dependencyChanged",
        #                 {"uri": uri},
        #             )

        robot_workspace = RobotWorkspace(
            root_uri,
            fs_observer,
            workspace_folders,
            libspec_manager=self.libspec_manager,
            index_workspace=self._index_workspace,
            collect_tests=self._collect_tests,
            endpoint=self._endpoint,
        )

        return robot_workspace

    def m_lint(self, doc_uri):
        error = self._compute_min_version_error((3, 2))
        if error is not None:
            from robocorp_ls_core.lsp import Error

            return [Error(error, (0, 0), (1, 0)).to_lsp_diagnostic()]

        func = partial(self._threaded_lint, doc_uri)
        func = require_monitor(func)
        return func

    def _threaded_lint(self, doc_uri, monitor: IMonitor):
        from robocorp_ls_core.jsonrpc.exceptions import JsonRpcRequestCancelled
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_ROBOCOP_ENABLED,
        )
        from robocorp_ls_core import uris
        from robocorp_ls_core.lsp import Error
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_LINT_ENABLED,
        )

        try:
            from robotframework_ls.impl.ast_utils import collect_errors
            from robotframework_ls.impl import code_analysis
            import os.path

            log.debug("Lint: starting (in thread).")

            completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
            if completion_context is None:
                return []

            config = completion_context.config
            robocop_enabled = config is None or config.get_setting(
                OPTION_ROBOT_LINT_ROBOCOP_ENABLED, bool, False
            )

            ast = completion_context.get_ast()
            source = completion_context.doc.source
            monitor.check_cancelled()
            errors = collect_errors(ast)
            log.debug("Collected AST errors (in thread): %s", len(errors))
            monitor.check_cancelled()

            lint_ls_enabled = config is None or config.get_setting(
                OPTION_ROBOT_LINT_ENABLED, bool, True
            )
            if lint_ls_enabled:
                analysis_errors = code_analysis.collect_analysis_errors(
                    completion_context
                )
                monitor.check_cancelled()
                log.debug(
                    "Collected analysis errors (in thread): %s", len(analysis_errors)
                )
                errors.extend(analysis_errors)
            else:
                log.debug("Language server linting disabled.")

            lsp_diagnostics = [error.to_lsp_diagnostic() for error in errors]

            try:
                if robocop_enabled:
                    from robocorp_ls_core.robocop_wrapper import (
                        collect_robocop_diagnostics,
                    )

                    workspace = completion_context.workspace
                    if workspace is not None:
                        project_root = workspace.root_path
                    else:
                        project_root = os.path.abspath(".")

                    monitor.check_cancelled()
                    lsp_diagnostics.extend(
                        collect_robocop_diagnostics(
                            project_root, ast, uris.to_fs_path(doc_uri), source
                        )
                    )
            except Exception as e:
                log.exception(
                    "Error collecting Robocop errors (possibly an unsupported Robocop version is installed)."
                )
                lsp_diagnostics.append(
                    Error(
                        f"Error collecting Robocop errors: {e}", (0, 0), (1, 0)
                    ).to_lsp_diagnostic()
                )

            return lsp_diagnostics
        except JsonRpcRequestCancelled:
            raise JsonRpcRequestCancelled("Lint cancelled (inside lint)")
        except Exception as e:
            log.exception("Error collecting errors.")
            ret = [
                Error(
                    f"Error collecting Robocop errors: {e}", (0, 0), (1, 0)
                ).to_lsp_diagnostic()
            ]
            return ret

    def m_resolve_completion_item(
        self,
        completion_item: CompletionItemTypedDict,
        monaco=False,
    ) -> CompletionItemTypedDict:
        # Note: don't put it in a thread as it should be cheap to do in the main thread.
        use_ctx = None
        with self._completion_contexts_saved_lock:
            data = completion_item.get("data")
            if data and isinstance(data, dict):
                ctx_id = data.get("ctx")
                if ctx_id is not None:
                    for ctx in self._completion_contexts_saved:
                        if id(ctx) == ctx_id:
                            use_ctx = ctx
                            break

        if use_ctx is not None:
            use_ctx.resolve_completion_item(data, completion_item, monaco=monaco)

        return completion_item

    def m_complete_all(self, doc_uri, line, col):
        func = partial(self._threaded_complete_all, doc_uri, line, col)
        func = require_monitor(func)
        return func

    def _threaded_complete_all(self, doc_uri, line, col, monitor: IMonitor):
        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return []

        from robotframework_ls.impl import snippets_completions

        completions = snippets_completions.complete(completion_context)
        monitor.check_cancelled()
        completions.extend(self._complete_from_completion_context(completion_context))
        return completions

    def _complete_from_completion_context(self, completion_context):
        with self._completion_contexts_saved_lock:
            # Keep up to 3 saved contexts there (used to resolve
            # completion items afterwards).
            while len(self._completion_contexts_saved) > 2:
                self._completion_contexts_saved.popleft()
            self._completion_contexts_saved.append(completion_context)

        return complete_all(completion_context)

    def m_section_name_complete(self, doc_uri, line, col):
        from robotframework_ls.impl import section_name_completions

        completion_context = self._create_completion_context(doc_uri, line, col, None)
        if completion_context is None:
            return []

        return section_name_completions.complete(completion_context)

    def m_keyword_complete(self, doc_uri, line, col):
        from robotframework_ls.impl import keyword_completions

        completion_context = self._create_completion_context(doc_uri, line, col, None)
        if completion_context is None:
            return []
        return keyword_completions.complete(completion_context)

    def m_find_definition(self, doc_uri, line, col):
        func = partial(self._threaded_find_definition, doc_uri, line, col)
        func = require_monitor(func)
        return func

    def _threaded_find_definition(self, doc_uri, line, col, monitor) -> Optional[list]:
        from robotframework_ls.impl.find_definition import find_definition_extended
        import os.path
        from robocorp_ls_core.lsp import Location, Range
        from robocorp_ls_core import uris
        from robocorp_ls_core.lsp import LocationLink

        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None
        definition_info = find_definition_extended(completion_context)

        if definition_info is None:
            return []

        origin_selection_range: Optional[
            RangeTypedDict
        ] = definition_info.origin_selection_range

        ret = []
        definition: IDefinition
        for definition in definition_info.definitions:
            if not definition.source:
                log.info("Found definition with empty source (%s).", definition)
                continue

            if not os.path.exists(definition.source):
                log.info(
                    "Found definition: %s (but source does not exist).", definition
                )
                continue

            lineno = definition.lineno
            if lineno is None or lineno < 0:
                lineno = 0

            end_lineno = definition.end_lineno
            if end_lineno is None or end_lineno < 0:
                end_lineno = 0

            col_offset = definition.col_offset
            end_col_offset = definition.end_col_offset

            if origin_selection_range is None:
                ret.append(
                    Location(
                        uris.from_fs_path(definition.source),
                        Range((lineno, col_offset), (end_lineno, end_col_offset)),
                    ).to_dict()
                )
            else:
                target_range = Range((lineno, col_offset), (end_lineno, end_col_offset))
                target_selection_range = target_range

                scope_lineno = definition.scope_lineno
                if scope_lineno is not None:
                    scope_col_offset = definition.scope_col_offset
                    if scope_col_offset is not None:
                        scope_end_lineno = definition.scope_end_lineno
                        scope_end_col_offset = definition.scope_end_col_offset
                        target_range = Range(
                            (scope_lineno, scope_col_offset),
                            (scope_end_lineno, scope_end_col_offset),
                        )

                ret.append(
                    LocationLink(
                        origin_selection_range,
                        uris.from_fs_path(definition.source),
                        target_range=target_range,
                        target_selection_range=target_selection_range,
                    ).to_dict()
                )
        return ret

    def m_code_format(self, text_document, options):
        func = partial(self._threaded_code_format, text_document, options)
        func = require_monitor(func)
        return func

    def _threaded_code_format(self, text_document, options, monitor: IMonitor):
        from robotframework_ls.impl.formatting import create_text_edit_from_diff
        from robocorp_ls_core.lsp import TextDocumentItem
        import os.path
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_CODE_FORMATTER,
        )
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY,
        )
        from robotframework_ls.impl.robot_lsp_constants import (
            OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY,
        )

        text_document_item = TextDocumentItem(**text_document)
        text = text_document_item.text
        if not text:
            completion_context = self._create_completion_context(
                text_document_item.uri, 0, 0, monitor
            )
            if completion_context is None:
                return []
            text = completion_context.doc.source

        if not text:
            return []

        if not self._check_and_log_rf_dependency_version():
            return []

        if options is None:
            options = {}
        tab_size = options.get("tabSize", 4)

        # Default for now is the builtin. This will probably be changed in the future.
        formatter = self._config.get_setting(
            OPTION_ROBOT_CODE_FORMATTER, str, OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY
        )
        if formatter not in (
            OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY,
            OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY,
        ):
            log.critical(
                f"Code formatter invalid: {formatter}. Please select one of: {OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY}, {OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY}."
            )
            return []

        if formatter == OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY:
            try:
                from robot.tidy import Tidy
            except ImportError:
                # It's not available in newer versions of RobotFramework.
                from robotframework_ls.impl.robot_version import get_robot_major_version

                if get_robot_major_version() >= 5:
                    formatter = OPTION_ROBOT_CODE_FORMATTER_ROBOTIDY

        if formatter == OPTION_ROBOT_CODE_FORMATTER_BUILTIN_TIDY:
            from robotframework_ls.impl.formatting import robot_source_format

            new_contents = robot_source_format(text, space_count=tab_size)

        else:
            error = self._compute_min_version_error((4, 0))
            if error is not None:
                log.critical(
                    f"To use the robotidy formatter, at least Robot Framework 4 is needed. {error}"
                )
                return []

            from robocorp_ls_core.robotidy_wrapper import robot_tidy_source_format

            ast = completion_context.get_ast()
            path = completion_context.doc.path
            dirname = "."
            try:
                os.stat(path)
            except:
                # It doesn't exist
                ws = self._workspace
                if ws is not None:
                    dirname = ws.root_path
            else:
                dirname = os.path.dirname(path)

            try:
                new_contents = robot_tidy_source_format(ast, dirname)
            except ImportError:
                log.critical(
                    "Unable to code-format because robotidy could not be imported."
                )
                return []

        if new_contents is None or new_contents == text:
            return []
        return [x.to_dict() for x in create_text_edit_from_diff(text, new_contents)]

    def _create_completion_context(
        self, doc_uri, line, col, monitor: Optional[IMonitor]
    ):
        from robotframework_ls.impl.completion_context import CompletionContext

        if not self._check_and_log_rf_dependency_version():
            return None
        workspace = self.workspace
        if not workspace:
            log.info("Workspace still not initialized.")
            return None

        document = workspace.get_document(doc_uri, accept_from_file=True)
        if document is None:
            log.info("Unable to get document for uri: %s.", doc_uri)
            return None
        return CompletionContext(
            document,
            line,
            col,
            workspace=workspace,
            config=self.config,
            monitor=monitor,
            variables_from_arguments_files_loader=self._variables_from_arguments_files_loader,
        )

    def m_signature_help(self, doc_uri: str, line: int, col: int):
        func = partial(self._threaded_signature_help, doc_uri, line, col)
        func = require_monitor(func)
        return func

    def _threaded_signature_help(
        self, doc_uri: str, line: int, col: int, monitor: IMonitor
    ) -> Optional[dict]:
        from robotframework_ls.impl.signature_help import signature_help

        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None

        return signature_help(completion_context)

    def m_folding_range(self, doc_uri: str):
        func = partial(self._threaded_folding_range, doc_uri)
        func = require_monitor(func)
        return func

    def _threaded_folding_range(
        self, doc_uri: str, monitor: IMonitor
    ) -> List[FoldingRangeTypedDict]:
        from robotframework_ls.impl.folding_range import folding_range

        completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
        if completion_context is None:
            return []

        return folding_range(completion_context)

    def m_code_lens(self, doc_uri: str):
        func = partial(self._threaded_code_lens, doc_uri)
        func = require_monitor(func)
        return func

    def _threaded_code_lens(
        self, doc_uri: str, monitor: IMonitor
    ) -> List[CodeLensTypedDict]:
        from robotframework_ls.impl.code_lens import code_lens

        completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
        if completion_context is None:
            return []

        return code_lens(completion_context)

    def m_resolve_code_lens(self, **code_lens: CodeLensTypedDict):
        func = partial(self._threaded_resolve_code_lens, code_lens)
        func = require_monitor(func)
        return func

    def _threaded_resolve_code_lens(
        self, code_lens: CodeLensTypedDict, monitor: IMonitor
    ) -> CodeLensTypedDict:
        from robotframework_ls.impl.code_lens import code_lens_resolve

        data = code_lens.get("data")
        if not isinstance(data, dict):
            return code_lens

        doc_uri = data.get("uri")
        completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
        if completion_context is None:
            return code_lens

        return code_lens_resolve(completion_context, code_lens)

    def m_document_symbol(self, doc_uri: str):
        func = partial(self._threaded_document_symbol, doc_uri)
        func = require_monitor(func)
        return func

    def _threaded_document_symbol(
        self, doc_uri: str, monitor: IMonitor
    ) -> List[DocumentSymbolTypedDict]:
        from robotframework_ls.impl.document_symbol import document_symbol

        completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
        if completion_context is None:
            return []

        return document_symbol(completion_context)

    def m_list_tests(self, doc_uri: str):
        func = partial(self._threaded_list_tests, doc_uri)
        func = require_monitor(func)
        return func

    def _threaded_list_tests(
        self, doc_uri: str, monitor: IMonitor
    ) -> List[ITestInfoTypedDict]:
        from robotframework_ls.impl.code_lens import list_tests

        completion_context = self._create_completion_context(doc_uri, 0, 0, monitor)
        if completion_context is None:
            return []

        return list_tests(completion_context)

    def m_evaluatable_expression(self, doc_uri: str, position: PositionTypedDict):
        func = partial(self._threaded_evaluatable_expression, doc_uri, position)
        func = require_monitor(func)
        return func

    def _threaded_evaluatable_expression(
        self, doc_uri: str, position: PositionTypedDict, monitor: IMonitor
    ) -> Optional[EvaluatableExpressionTypedDict]:
        from robotframework_ls.impl.provide_evaluatable_expression import (
            provide_evaluatable_expression,
        )

        line = position["line"]
        col = position["character"]
        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None

        return provide_evaluatable_expression(completion_context)

    def m_wait_for_full_test_collection(self):
        func = partial(self._threaded_wait_for_full_test_collection)
        func = require_monitor(func)
        return func

    def _threaded_wait_for_full_test_collection(self, monitor: IMonitor) -> bool:
        from robotframework_ls.impl.robot_workspace import RobotWorkspace

        workspace = self.workspace
        if not workspace:
            log.info("Workspace still not initialized.")
            return False
        ws = typing.cast(RobotWorkspace, workspace)
        workspace_indexer = ws.workspace_indexer
        if not workspace_indexer:
            raise RuntimeError("WorkspaceIndexer not available (None).")
        workspace_indexer.wait_for_full_test_collection()
        return True

    def m_hover(self, doc_uri: str, line: int, col: int):
        func = partial(self._threaded_hover, doc_uri, line, col)
        func = require_monitor(func)
        return func

    def _threaded_hover(
        self, doc_uri: str, line, col, monitor: IMonitor
    ) -> Optional[HoverTypedDict]:
        from robotframework_ls.impl.hover import hover

        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None

        return hover(completion_context)

    def m_document_highlight(self, doc_uri: str, line: int, col: int):
        func = partial(self._threaded_document_highlight, doc_uri, line, col)
        func = require_monitor(func)
        return func

    def _threaded_document_highlight(
        self, doc_uri: str, line, col, monitor: IMonitor
    ) -> Optional[List[DocumentHighlightTypedDict]]:

        from robotframework_ls.impl.doc_highlight import doc_highlight

        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None

        return doc_highlight(completion_context)

    def m_references(
        self, doc_uri: str, line: int, col: int, include_declaration: bool
    ):
        func = partial(
            self._threaded_references, doc_uri, line, col, include_declaration
        )
        func = require_monitor(func)
        return func

    def _threaded_references(
        self, doc_uri: str, line, col, include_declaration: bool, monitor: IMonitor
    ) -> Optional[List[LocationTypedDict]]:
        from robotframework_ls.impl.references import references

        completion_context = self._create_completion_context(
            doc_uri, line, col, monitor
        )
        if completion_context is None:
            return None

        return references(completion_context, include_declaration=include_declaration)

    def m_workspace_symbols(self, query: Optional[str] = None):
        func = partial(self._threaded_workspace_symbols, query)
        func = require_monitor(func)
        return func

    def _threaded_workspace_symbols(
        self, query: Optional[str], monitor: IMonitor
    ) -> Optional[List[SymbolInformationTypedDict]]:
        from robotframework_ls.impl.workspace_symbols import workspace_symbols
        from robotframework_ls.impl.completion_context import BaseContext
        from robotframework_ls.impl.protocols import IRobotWorkspace
        from typing import cast

        workspace = self._workspace
        if not workspace:
            return []

        robot_workspace = cast(IRobotWorkspace, workspace)

        return workspace_symbols(
            query,
            BaseContext(workspace=robot_workspace, config=self.config, monitor=monitor),
        )

    def m_text_document__semantic_tokens__range(self, textDocument=None, range=None):
        raise RuntimeError("Not currently implemented!")

    def m_text_document__semantic_tokens__full(self, textDocument=None):
        func = partial(self.threaded_semantic_tokens_full, textDocument=textDocument)
        func = require_monitor(func)
        return func

    def threaded_semantic_tokens_full(
        self, textDocument: TextDocumentTypedDict, monitor: Optional[IMonitor] = None
    ):
        from robotframework_ls.impl.semantic_tokens import semantic_tokens_full

        doc_uri = textDocument["uri"]
        context = self._create_completion_context(doc_uri, -1, -1, monitor)
        if context is None:
            return {"resultId": None, "data": []}
        return {"resultId": None, "data": semantic_tokens_full(context)}

    def m_monaco_completions_from_code_full(
        self,
        prefix: str = "",
        full_code: str = "",
        position=PositionTypedDict,
        uri: str = "",
        indent: str = "",
    ):
        func = partial(
            self.threaded_monaco_completions_from_code_full,
            prefix=prefix,
            full_code=full_code,
            position=position,
            uri=uri,
            indent=indent,
        )
        func = require_monitor(func)
        return func

    def m_monaco_resolve_completion(
        self,
        completion_item: CompletionItemTypedDict,
    ):
        return self.m_resolve_completion_item(completion_item, monaco=True)

    def threaded_monaco_completions_from_code_full(
        self,
        prefix: str,
        full_code: str,
        position: PositionTypedDict,
        uri: str,
        indent: str,
        monitor: Optional[IMonitor] = None,
    ):
        from robotframework_ls.impl.robot_workspace import RobotDocument
        from robotframework_ls.impl.completion_context import CompletionContext
        from robocorp_ls_core.workspace import Document
        from robotframework_ls.impl import section_completions
        from robotframework_ls.impl import snippets_completions
        from robotframework_ls.server_api.monaco_conversions import (
            convert_to_monaco_completion,
        )
        from robotframework_ls.impl.completion_context import CompletionType

        d = Document(uri, prefix)
        last_line, _last_col = d.get_last_line_col()
        line = last_line + position["line"]

        col = position["character"]
        col += len(indent)

        document = RobotDocument(uri, full_code)
        completion_context = CompletionContext(
            document,
            line,
            col,
            config=self.config,
            monitor=monitor,
            workspace=self.workspace,
        )
        completion_context.type = CompletionType.shell
        completions = self._complete_from_completion_context(completion_context)
        completions.extend(section_completions.complete(completion_context))
        completions.extend(snippets_completions.complete(completion_context))

        return {
            "suggestions": [
                convert_to_monaco_completion(
                    c, line_delta=last_line, col_delta=len(indent), uri=uri
                )
                for c in completions
            ]
        }

    def m_semantic_tokens_from_code_full(
        self, prefix: str = "", full_code: str = "", indent: str = "", uri: str = ""
    ):
        func = partial(
            self.threaded_semantic_tokens_from_code_full,
            prefix=prefix,
            full_code=full_code,
            indent=indent,
            uri=uri,
        )
        func = require_monitor(func)
        return func

    def threaded_semantic_tokens_from_code_full(
        self,
        prefix: str,
        full_code: str,
        indent: str,
        uri: str,
        monitor: Optional[IMonitor] = None,
    ):
        from robotframework_ls.impl.semantic_tokens import semantic_tokens_full
        from robotframework_ls.impl.completion_context import CompletionContext

        try:
            from robotframework_ls.impl.robot_workspace import RobotDocument

            document = RobotDocument(uri, full_code)
            completion_context = CompletionContext(
                document,
                config=self.config,
                monitor=monitor,
                workspace=self.workspace,
            )

            data = semantic_tokens_full(completion_context)
            if not prefix:
                return {"resultId": None, "data": data}

            # We have to exclude the prefix from the coloring...

            # debug info...
            # import io
            # from robotframework_ls.impl.semantic_tokens import decode_semantic_tokens
            # stream = io.StringIO()
            # decode_semantic_tokens(data, doc, stream)
            # found = stream.getvalue()

            prefix_doc = RobotDocument("")
            prefix_doc.source = prefix
            last_line, last_col = prefix_doc.get_last_line_col()

            # Now we have the data from the full code, but we need to remove whatever
            # we have in the prefix from the result...
            ints_iter = iter(data)
            line = 0
            col = 0
            new_data = []
            indent_len = len(indent)
            while True:
                try:
                    line_delta = next(ints_iter)
                except StopIteration:
                    break
                col_delta = next(ints_iter)
                token_len = next(ints_iter)
                token_type = next(ints_iter)
                token_modifier = next(ints_iter)
                line += line_delta
                if line_delta == 0:
                    col += col_delta
                else:
                    col = col_delta

                if line >= last_line:
                    new_data.append(line - last_line)
                    new_data.append(col_delta - indent_len)
                    new_data.append(token_len)
                    new_data.append(token_type)
                    new_data.append(token_modifier)

                    # Ok, now, we have to add the indent_len to all the
                    # next lines
                    while True:
                        try:
                            line_delta = next(ints_iter)
                        except StopIteration:
                            break
                        col_delta = next(ints_iter)
                        token_len = next(ints_iter)
                        token_type = next(ints_iter)
                        token_modifier = next(ints_iter)

                        new_data.append(line_delta)
                        if line_delta > 0:
                            new_data.append(col_delta - indent_len)
                        else:
                            new_data.append(col_delta)
                        new_data.append(token_len)
                        new_data.append(token_type)
                        new_data.append(token_modifier)

                    break

                # Approach changed so that we always have a new line
                # i.e.:
                # \n<indent><code>
                #
                # so, the condition below no longer applies.
                # elif line == last_line and col >= last_col:
                #     new_data.append(0)
                #     new_data.append(col - last_col)
                #     new_data.append(token_len)
                #     new_data.append(token_type)
                #     new_data.append(token_modifier)
                #     new_data.extend(ints_iter)
                #     break

            # debug info...
            # temp_stream = io.StringIO()
            # temp_doc = RobotDocument("")
            # temp_doc.source = full_code[len(prefix) :]
            # decode_semantic_tokens(new_data, temp_doc, temp_stream)
            # temp_found = temp_stream.getvalue()

            return {"resultId": None, "data": new_data}
        except:
            log.exception("Error computing semantic tokens from code.")
            return {"resultId": None, "data": []}

    def m_shutdown(self, **_kwargs):
        PythonLanguageServer.m_shutdown(self, **_kwargs)
        self.libspec_manager.dispose()
        workspace = self._workspace
        if workspace is not None:
            workspace.dispose()

    def m_exit(self, **_kwargs):
        PythonLanguageServer.m_exit(self, **_kwargs)
        self.libspec_manager.dispose()
        workspace = self._workspace
        if workspace is not None:
            workspace.dispose()



import sys
import threading
from _pydev_bundle import pydev_log


def check():
    with pydev_log.log_context(3, sys.stderr):
        assert hasattr(sys, 'gettotalrefcount')
        import pydevd_tracing

        proceed1 = threading.Event()
        proceed2 = threading.Event()

        class SomeThread(threading.Thread):

            def run(self):
                proceed1.set()
                proceed2.wait()

        t = SomeThread()
        t.start()
        proceed1.wait()
        try:

            def some_func(frame, event, arg):
                return some_func

            pydevd_tracing.set_trace_to_threads(some_func)
        finally:
            proceed2.set()

        lib = pydevd_tracing._load_python_helper_lib()
        assert lib is None
        print('Finished OK')


if __name__ == '__main__':
    check()



#!/usr/bin/env python
from gevent import monkey, sleep, threading as gevent_threading
import sys

if 'remote' in sys.argv:
    import pydevd
    port = int(sys.argv[1])
    print('before pydevd.settrace')
    pydevd.settrace(host=('' if 'as-server' in sys.argv else '127.0.0.1'), port=port, suspend=False)
    print('after pydevd.settrace')

monkey.patch_all()
import threading

called = []


class MyGreenThread2(threading.Thread):

    def run(self):
        for _i in range(3):
            sleep()


class MyGreenletThread(threading.Thread):

    def run(self):
        for _i in range(5):
            called.append(self.name)  # break here
            t1 = MyGreenThread2()
            t1.start()
            sleep()


if __name__ == '__main__':
    t1 = MyGreenletThread()
    t1.name = 't1'
    t2 = MyGreenletThread()
    t2.name = 't2'

    if hasattr(gevent_threading, 'Thread'):
        # Only available in newer versions of gevent.
        assert isinstance(t1, gevent_threading.Thread)
        assert isinstance(t2, gevent_threading.Thread)

    t1.start()
    t2.start()

    for t1 in (t1, t2):
        t1.join()

    # With gevent it's always the same (gevent coroutine support makes thread
    # switching serial).
    expected = ['t1', 't2', 't1', 't2', 't1', 't2', 't1', 't2', 't1', 't2']
    if called != expected:
        raise AssertionError("Expected:\n%s\nFound:\n%s" % (expected, called))
    print('TEST SUCEEDED')



import subprocess
import sys
import os
from contextlib import contextmanager
env = os.environ.copy()
pythonpath = env.get('PYTHONPATH', '')
env['PYTHONPATH'] = os.path.dirname(__file__)


def call():
    print("called")  # break 1 here


@contextmanager
def dummy_context_manager():
    yield


if __name__ == '__main__':
    if 'apply-multiprocessing-patch' in sys.argv:
        context_manager = dummy_context_manager  # It's applied by default.
    else:
        import pydevd
        context_manager = pydevd.skip_subprocess_arg_patch

    with context_manager():
        p = subprocess.Popen(
            [sys.executable, '-c', 'import _debugger_case_no_subprocess_patching;_debugger_case_no_subprocess_patching.call()'],
            stdout=subprocess.PIPE,
            env=env,
        )

    stdout, stderr = p.communicate()
    assert b'called' in stdout, 'Did not find b"called" in: %s' % (stdout,)
    print('TEST SUCEEDED!')  # break 2 here



import threading
import unittest
import sys
import pydevconsole
from _pydev_bundle.pydev_imports import xmlrpclib, SimpleXMLRPCServer
from _pydevd_bundle import pydevd_io

try:
    raw_input
    raw_input_name = 'raw_input'
except NameError:
    raw_input_name = 'input'


#=======================================================================================================================
# Test
#=======================================================================================================================
class Test(unittest.TestCase):

    def test_console_hello(self):
        self.original_stdout = sys.stdout
        sys.stdout = pydevd_io.IOBuf()
        try:
            sys.stdout.encoding = sys.stdin.encoding
        except AttributeError:
            # In Python 3 encoding is not writable (whereas in Python 2 it doesn't exist).
            pass

        try:
            client_port, _server_port = self.get_free_addresses()
            client_thread = self.start_client_thread(client_port)  # @UnusedVariable
            import time
            time.sleep(.3)  # let's give it some time to start the threads

            from _pydev_bundle import pydev_localhost
            interpreter = pydevconsole.InterpreterInterface(pydev_localhost.get_localhost(), client_port, threading.current_thread())

            (result,) = interpreter.hello("Hello pydevconsole")
            self.assertEqual(result, "Hello eclipse")
        finally:
            sys.stdout = self.original_stdout

    def test_console_requests(self):
        self.original_stdout = sys.stdout
        sys.stdout = pydevd_io.IOBuf()

        try:
            client_port, _server_port = self.get_free_addresses()
            client_thread = self.start_client_thread(client_port)  # @UnusedVariable
            import time
            time.sleep(.3)  # let's give it some time to start the threads

            from _pydev_bundle import pydev_localhost
            from _pydev_bundle.pydev_console_utils import CodeFragment

            interpreter = pydevconsole.InterpreterInterface(pydev_localhost.get_localhost(), client_port, threading.current_thread())
            sys.stdout = pydevd_io.IOBuf()
            interpreter.add_exec(CodeFragment('class Foo:\n    CONSTANT=1\n'))
            interpreter.add_exec(CodeFragment('foo=Foo()'))
            interpreter.add_exec(CodeFragment('foo.__doc__=None'))
            interpreter.add_exec(CodeFragment('val = %s()' % (raw_input_name,)))
            interpreter.add_exec(CodeFragment('50'))
            interpreter.add_exec(CodeFragment('print (val)'))
            found = sys.stdout.getvalue().split()
            try:
                self.assertEqual(['50', 'input_request'], found)
            except:
                try:
                    self.assertEqual(['input_request'], found)  # IPython
                except:
                    self.assertEqual([u'50', u'input_request'], found[1:])  # IPython 5.1
                    self.assertTrue(found[0].startswith(u'Out'))

            comps = interpreter.getCompletions('foo.', 'foo.')
            self.assertTrue(
                ('CONSTANT', '', '', '3') in comps or ('CONSTANT', '', '', '4') in comps, \
                'Found: %s' % comps
            )

            comps = interpreter.getCompletions('"".', '"".')
            self.assertTrue(
                ('__add__', 'x.__add__(y) <==> x+y', '', '3') in comps or
                ('__add__', '', '', '4') in comps or
                ('__add__', 'x.__add__(y) <==> x+y\r\nx.__add__(y) <==> x+y', '()', '2') in comps or
                ('__add__', 'x.\n__add__(y) <==> x+yx.\n__add__(y) <==> x+y', '()', '2'),
                'Did not find __add__ in : %s' % (comps,)
            )

            completions = interpreter.getCompletions('', '')
            for c in completions:
                if c[0] == 'AssertionError':
                    break
            else:
                self.fail('Could not find AssertionError')

            completions = interpreter.getCompletions('Assert', 'Assert')
            for c in completions:
                if c[0] == 'RuntimeError':
                    self.fail('Did not expect to find RuntimeError there')

            assert ('__doc__', None, '', '3') not in interpreter.getCompletions('foo.CO', 'foo.')

            comps = interpreter.getCompletions('va', 'va')
            assert ('val', '', '', '3') in comps or ('val', '', '', '4') in comps

            interpreter.add_exec(CodeFragment('s = "mystring"'))

            desc = interpreter.getDescription('val')
            self.assertTrue(desc.find('str(object) -> string') >= 0 or
                         desc == "'input_request'" or
                         desc.find('str(string[, encoding[, errors]]) -> str') >= 0 or
                         desc.find('str(Char* value)') >= 0 or
                         desc.find('str(object=\'\') -> string') >= 0 or
                         desc.find('str(value: Char*)') >= 0 or
                         desc.find('str(object=\'\') -> str') >= 0 or
                         desc.find('The most base type') >= 0  # Jython 2.7 is providing this :P
                         ,
                         'Could not find what was needed in %s' % desc)

            desc = interpreter.getDescription('val.join')
            self.assertTrue(desc.find('S.join(sequence) -> string') >= 0 or
                         desc.find('S.join(sequence) -> str') >= 0 or
                         desc.find('S.join(iterable) -> string') >= 0 or
                         desc == "<builtin method 'join'>"  or
                         desc == "<built-in method join of str object>" or
                         desc.find('str join(str self, list sequence)') >= 0 or
                         desc.find('S.join(iterable) -> str') >= 0 or
                         desc.find('join(self: str, sequence: list) -> str') >= 0 or
                         desc.find('Concatenate any number of strings.') >= 0 or
                         desc.find('bound method str.join') >= 0,  # PyPy
                         "Could not recognize: %s" % (desc,))
        finally:
            sys.stdout = self.original_stdout

    def start_client_thread(self, client_port):

        class ClientThread(threading.Thread):

            def __init__(self, client_port):
                threading.Thread.__init__(self)
                self.client_port = client_port

            def run(self):

                class HandleRequestInput:

                    def RequestInput(self):
                        client_thread.requested_input = True
                        return 'input_request'

                    def NotifyFinished(self, *args, **kwargs):
                        client_thread.notified_finished += 1
                        return 1

                handle_request_input = HandleRequestInput()

                from _pydev_bundle import pydev_localhost
                client_server = SimpleXMLRPCServer((pydev_localhost.get_localhost(), self.client_port), logRequests=False)
                client_server.register_function(handle_request_input.RequestInput)
                client_server.register_function(handle_request_input.NotifyFinished)
                client_server.serve_forever()

        client_thread = ClientThread(client_port)
        client_thread.requested_input = False
        client_thread.notified_finished = 0
        client_thread.daemon = True
        client_thread.start()
        return client_thread

    def start_debugger_server_thread(self, debugger_port, socket_code):

        class DebuggerServerThread(threading.Thread):

            def __init__(self, debugger_port, socket_code):
                threading.Thread.__init__(self)
                self.debugger_port = debugger_port
                self.socket_code = socket_code

            def run(self):
                import socket
                s = socket.socket()
                s.bind(('', debugger_port))
                s.listen(1)
                socket, unused_addr = s.accept()
                socket_code(socket)

        debugger_thread = DebuggerServerThread(debugger_port, socket_code)
        debugger_thread.daemon = True
        debugger_thread.start()
        return debugger_thread

    def get_free_addresses(self):
        from _pydev_bundle.pydev_localhost import get_socket_names
        socket_names = get_socket_names(2, True)
        port0 = socket_names[0][1]
        port1 = socket_names[1][1]

        assert port0 != port1
        assert port0 > 0
        assert port1 > 0

        return port0, port1

    def test_server(self):
        self.original_stdout = sys.stdout
        sys.stdout = pydevd_io.IOBuf()
        try:
            client_port, server_port = self.get_free_addresses()

            class ServerThread(threading.Thread):

                def __init__(self, client_port, server_port):
                    threading.Thread.__init__(self)
                    self.client_port = client_port
                    self.server_port = server_port

                def run(self):
                    from _pydev_bundle import pydev_localhost
                    pydevconsole.start_server(pydev_localhost.get_localhost(), self.server_port, self.client_port)

            server_thread = ServerThread(client_port, server_port)
            server_thread.daemon = True
            server_thread.start()

            client_thread = self.start_client_thread(client_port)  # @UnusedVariable

            import time
            time.sleep(.3)  # let's give it some time to start the threads
            sys.stdout = pydevd_io.IOBuf()

            from _pydev_bundle import pydev_localhost
            server = xmlrpclib.Server('http://%s:%s' % (pydev_localhost.get_localhost(), server_port))
            server.execLine('class Foo:')
            server.execLine('    pass')
            server.execLine('')
            server.execLine('foo = Foo()')
            server.execLine('a = %s()' % (raw_input_name,))
            server.execLine('print (a)')
            initial = time.time()
            while not client_thread.requested_input:
                if time.time() - initial > 2:
                    raise AssertionError('Did not get the return asked before the timeout.')
                time.sleep(.1)

            found = sys.stdout.getvalue()
            while ['input_request'] != found.split():
                found += sys.stdout.getvalue()
                if time.time() - initial > 2:
                    break
                time.sleep(.1)
            self.assertEqual(['input_request'], found.split())
        finally:
            sys.stdout = self.original_stdout




""" optional jupyter[lab]_lsp integration
"""


def spec_v1(app):
    """A spec loader for jupyter_lsp"""
    return {
        "robotframework_ls": dict(
            version=2,
            argv=["robotframework_ls"],
            languages=["robotframework"],
            mime_types=["text/x-robotframework"],
            urls=dict(
                home="https://github.com/robocorp/robotframework-lsp",
                issues="https://github.com/robocorp/robotframework-lsp/issues",
            ),
            install=dict(
                pip="pip install robotframework-lsp",
                conda="conda install -c conda-forge robotframework-lsp",
            ),
        )
    }



import asyncio

async def gen():
    for i in range(10):
        yield i

async def run():
    async for p in gen():
        print(p)

if __name__ == "__main__":
    loop = asyncio.get_event_loop_policy().get_event_loop()
    loop.run_until_complete(run())
    print('TEST SUCEEDED')


import sys
if __name__ == '__main__':

    a = [1, 2, 3, 4]
    b = [2, 7]

    assert any(x in a for x in b)

    def gen():
        yield 1

    for i in gen():
        pass

    def gen2():
        yield 2
        if sys.version_info[:2] < (3, 7):
            # On Python 3.7 onwards this will generate an unhandled exception, which
            # is not what we want.
            raise StopIteration()

    for i in gen2():
        pass

    print('TEST SUCEEDED!')



import os.path
import sys

IS_JYTHON = sys.platform.find('java') != -1

try:
    this_file_name = __file__
except NameError:
    # stupid jython. plain old __file__ isnt working for some reason
    import test_runfiles  # @UnresolvedImport - importing the module itself
    this_file_name = test_runfiles.__file__

desired_runfiles_path = os.path.normpath(os.path.dirname(this_file_name) + "/..")
sys.path.insert(0, desired_runfiles_path)

from _pydev_runfiles import pydev_runfiles_unittest
from _pydev_runfiles import pydev_runfiles_xml_rpc
from _pydevd_bundle import pydevd_io

# remove existing pydev_runfiles from modules (if any), so that we can be sure we have the correct version
if 'pydev_runfiles' in sys.modules:
    del sys.modules['pydev_runfiles']
if '_pydev_runfiles.pydev_runfiles' in sys.modules:
    del sys.modules['_pydev_runfiles.pydev_runfiles']

from _pydev_runfiles import pydev_runfiles
import unittest
import tempfile
import re

try:
    set
except:
    from sets import Set as set

# this is an early test because it requires the sys.path changed
orig_syspath = sys.path
a_file = pydev_runfiles.__file__
pydev_runfiles.PydevTestRunner(pydev_runfiles.Configuration(files_or_dirs=[a_file]))
file_dir = os.path.dirname(os.path.dirname(a_file))
assert file_dir in sys.path
sys.path = orig_syspath[:]

# remove it so that we leave it ok for other tests
sys.path.remove(desired_runfiles_path)


class RunfilesTest(unittest.TestCase):

    def _setup_scenario(
        self,
        path,
        include_tests=None,
        tests=None,
        files_to_tests=None,
        exclude_files=None,
        exclude_tests=None,
        include_files=None,
        ):
        self.MyTestRunner = pydev_runfiles.PydevTestRunner(
            pydev_runfiles.Configuration(
                files_or_dirs=path,
                include_tests=include_tests,
                verbosity=1,
                tests=tests,
                files_to_tests=files_to_tests,
                exclude_files=exclude_files,
                exclude_tests=exclude_tests,
                include_files=include_files,
            )
        )
        self.files = self.MyTestRunner.find_import_files()
        self.modules = self.MyTestRunner.find_modules_from_files(self.files)
        self.all_tests = self.MyTestRunner.find_tests_from_modules(self.modules)
        self.filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)

    def setUp(self):
        self.file_dir = [os.path.abspath(os.path.join(desired_runfiles_path, 'tests_runfiles/samples'))]
        self._setup_scenario(self.file_dir, None)

    def test_suite_used(self):
        for suite in self.all_tests + self.filtered_tests:
            self.assertTrue(isinstance(suite, pydev_runfiles_unittest.PydevTestSuite))

    def test_parse_cmdline(self):
        sys.argv = "pydev_runfiles.py ./".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual([sys.argv[1]], configuration.files_or_dirs)
        self.assertEqual(2, configuration.verbosity)  # default value
        self.assertEqual(None, configuration.include_tests)  # default value

        sys.argv = "pydev_runfiles.py ../images c:/temp".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual(sys.argv[1:3], configuration.files_or_dirs)
        self.assertEqual(2, configuration.verbosity)

        sys.argv = "pydev_runfiles.py --verbosity 3 ../junk c:/asdf ".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual(sys.argv[3:], configuration.files_or_dirs)
        self.assertEqual(int(sys.argv[2]), configuration.verbosity)

        sys.argv = "pydev_runfiles.py --include_tests test_def ./".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual([sys.argv[-1]], configuration.files_or_dirs)
        self.assertEqual([sys.argv[2]], configuration.include_tests)

        sys.argv = "pydev_runfiles.py --include_tests Abc.test_def,Mod.test_abc c:/junk/".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual([sys.argv[-1]], configuration.files_or_dirs)
        self.assertEqual(sys.argv[2].split(','), configuration.include_tests)

        sys.argv = ('C:\\eclipse-SDK-3.2-win32\\eclipse\\plugins\\org.python.pydev.debug_1.2.2\\pysrc\\pydev_runfiles.py ' +
                    '--verbosity 1 ' +
                    'C:\\workspace_eclipse\\fronttpa\\tests\\gui_tests\\calendar_popup_control_test.py ').split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual([sys.argv[-1]], configuration.files_or_dirs)
        self.assertEqual(1, configuration.verbosity)

        sys.argv = "pydev_runfiles.py --verbosity 1 --include_tests Mod.test_abc c:/junk/ ./".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual(sys.argv[5:], configuration.files_or_dirs)
        self.assertEqual(int(sys.argv[2]), configuration.verbosity)
        self.assertEqual([sys.argv[4]], configuration.include_tests)

        sys.argv = "pydev_runfiles.py --exclude_files=*.txt,a*.py".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual(['*.txt', 'a*.py'], configuration.exclude_files)

        sys.argv = "pydev_runfiles.py --exclude_tests=*__todo,test*bar".split()
        configuration = pydev_runfiles.parse_cmdline()
        self.assertEqual(['*__todo', 'test*bar'], configuration.exclude_tests)

    def test___adjust_python_path_works_for_directories(self):
        orig_syspath = sys.path
        tempdir = tempfile.gettempdir()
        pydev_runfiles.PydevTestRunner(pydev_runfiles.Configuration(files_or_dirs=[tempdir]))
        self.assertEqual(1, tempdir in sys.path)
        sys.path = orig_syspath[:]

    def test___is_valid_py_file(self):
        isvalid = self.MyTestRunner._PydevTestRunner__is_valid_py_file
        self.assertEqual(1, isvalid("test.py"))
        self.assertEqual(0, isvalid("asdf.pyc"))
        self.assertEqual(0, isvalid("__init__.py"))
        self.assertEqual(0, isvalid("__init__.pyc"))
        self.assertEqual(1, isvalid("asdf asdf.pyw"))

    def test___unixify(self):
        unixify = self.MyTestRunner._PydevTestRunner__unixify
        self.assertEqual("c:/temp/junk/asdf.py", unixify("c:SEPtempSEPjunkSEPasdf.py".replace('SEP', os.sep)))

    def test___importify(self):
        importify = self.MyTestRunner._PydevTestRunner__importify
        self.assertEqual("temp.junk.asdf", importify("temp/junk/asdf.py"))
        self.assertEqual("asdf", importify("asdf.py"))
        self.assertEqual("abc.def.hgi", importify("abc/def/hgi"))

    def test_finding_a_file_from_file_system(self):
        test_file = "simple_test.py"
        self.MyTestRunner.files_or_dirs = [self.file_dir[0] + test_file]
        files = self.MyTestRunner.find_import_files()
        self.assertEqual(1, len(files))
        self.assertEqual(files[0], self.file_dir[0] + test_file)

    def test_finding_files_in_dir_from_file_system(self):
        self.assertEqual(1, len(self.files) > 0)
        for import_file in self.files:
            self.assertEqual(-1, import_file.find(".pyc"))
            self.assertEqual(-1, import_file.find("__init__.py"))
            self.assertEqual(-1, import_file.find("\\"))
            self.assertEqual(-1, import_file.find(".txt"))

    def test___get_module_from_str(self):
        my_importer = self.MyTestRunner._PydevTestRunner__get_module_from_str
        my_os_path = my_importer("os.path", True, 'unused')
        from os import path
        import os.path as path2
        self.assertEqual(path, my_os_path)
        self.assertEqual(path2, my_os_path)
        self.assertNotEqual(__import__("os.path"), my_os_path)
        self.assertNotEqual(__import__("os"), my_os_path)

    def test_finding_modules_from_import_strings(self):
        self.assertEqual(1, len(self.modules) > 0)

    def test_finding_tests_when_no_filter(self):
        # unittest.py will create a TestCase with 0 tests in it
        # since it just imports what is given
        self.assertEqual(1, len(self.all_tests) > 0)
        files_with_tests = [1 for t in self.all_tests if len(t._tests) > 0]
        self.assertNotEqual(len(self.files), len(files_with_tests))

    def count_suite(self, tests=None):
        total = 0
        for t in tests:
            total += t.countTestCases()
        return total

    def test_runfile_imports(self):
        from _pydev_runfiles import pydev_runfiles_coverage
        from _pydev_runfiles import pydev_runfiles_parallel_client
        from _pydev_runfiles import pydev_runfiles_parallel
        import pytest
        from _pydev_runfiles import pydev_runfiles_pytest2
        from _pydev_runfiles import pydev_runfiles_unittest
        from _pydev_runfiles import pydev_runfiles_xml_rpc
        from _pydev_runfiles import pydev_runfiles

    def test___match(self):
        matcher = self.MyTestRunner._PydevTestRunner__match
        self.assertEqual(1, matcher(None, "aname"))
        self.assertEqual(1, matcher([".*"], "aname"))
        self.assertEqual(0, matcher(["^x$"], "aname"))
        self.assertEqual(0, matcher(["abc"], "aname"))
        self.assertEqual(1, matcher(["abc", "123"], "123"))

    def test_finding_tests_from_modules_with_bad_filter_returns_0_tests(self):
        self._setup_scenario(self.file_dir, ["NO_TESTS_ARE_SURE_TO_HAVE_THIS_NAME"])
        self.assertEqual(0, self.count_suite(self.all_tests))

    def test_finding_test_with_unique_name_returns_1_test(self):
        self._setup_scenario(self.file_dir, include_tests=["test_i_am_a_unique_test_name"])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests))

    def test_finding_test_with_non_unique_name(self):
        self._setup_scenario(self.file_dir, include_tests=["test_non_unique_name"])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests) > 2)

    def test_finding_tests_with_regex_filters(self):
        self._setup_scenario(self.file_dir, include_tests=["test_non*"])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests) > 2)

        self._setup_scenario(self.file_dir, ["^$"])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(0, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, exclude_tests=["*"])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(0, self.count_suite(filtered_tests))

    def test_matching_tests(self):
        self._setup_scenario(self.file_dir, None, ['StillYetAnotherSampleTest'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, ['SampleTest.test_xxxxxx1'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, ['SampleTest'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(8, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, ['AnotherSampleTest.todo_not_tested'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(1, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, ['StillYetAnotherSampleTest', 'SampleTest.test_xxxxxx1'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(2, self.count_suite(filtered_tests))

        self._setup_scenario(self.file_dir, None, exclude_tests=['*'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(self.count_suite(filtered_tests), 0)

        self._setup_scenario(self.file_dir, None, exclude_tests=['*a*'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(self.count_suite(filtered_tests), 6)

        self.assertEqual(
            set(self.MyTestRunner.list_test_names(filtered_tests)),
            set(['test_1', 'test_2', 'test_xxxxxx1', 'test_xxxxxx2', 'test_xxxxxx3', 'test_xxxxxx4'])
        )

        self._setup_scenario(self.file_dir, None, exclude_tests=['*a*', '*x*'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        self.assertEqual(self.count_suite(filtered_tests), 2)

        self.assertEqual(
            set(self.MyTestRunner.list_test_names(filtered_tests)),
            set(['test_1', 'test_2'])
        )

        self._setup_scenario(self.file_dir, None, exclude_files=['simple_test.py'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        names = self.MyTestRunner.list_test_names(filtered_tests)
        self.assertTrue('test_xxxxxx1' not in names, 'Found: %s' % (names,))

        self.assertEqual(
            set(['test_abc', 'test_non_unique_name', 'test_non_unique_name', 'test_asdf2', 'test_i_am_a_unique_test_name', 'test_non_unique_name', 'test_blank']),
            set(names)
        )

        self._setup_scenario(self.file_dir, None, include_files=['simple3_test.py'])
        filtered_tests = self.MyTestRunner.filter_tests(self.all_tests)
        names = self.MyTestRunner.list_test_names(filtered_tests)
        self.assertTrue('test_xxxxxx1' not in names, 'Found: %s' % (names,))

        self.assertEqual(
            set(['test_non_unique_name']),
            set(names)
        )

    def test_xml_rpc_communication(self):
        import sys
        sys.path.insert(0, os.path.join(os.path.dirname(__file__), 'samples'))
        notifications = []

        class Server:

            def __init__(self, notifications):
                self.notifications = notifications

            def notifyConnected(self):
                # This method is called at the very start (in runfiles.py), and we do not check this here
                raise AssertionError('Should not be called from the run tests.')

            def notifyTestsCollected(self, number_of_tests):
                self.notifications.append(('notifyTestsCollected', number_of_tests))

            def notifyStartTest(self, file, test):
                pass

            def notifyTest(self, cond, captured_output, error_contents, file, test, time):
                try:
                    # I.e.: when marked as Binary in xml-rpc
                    captured_output = captured_output.data
                except:
                    pass
                try:
                    # I.e.: when marked as Binary in xml-rpc
                    error_contents = error_contents.data
                except:
                    pass
                if error_contents:
                    error_contents = error_contents.splitlines()[-1].strip()
                self.notifications.append(('notifyTest', cond, captured_output.strip(), error_contents, file, test))

            def notifyTestRunFinished(self, total_time):
                self.notifications.append(('notifyTestRunFinished',))

        server = Server(notifications)
        pydev_runfiles_xml_rpc.set_server(server)
        simple_test = os.path.join(self.file_dir[0], 'simple_test.py')
        simple_test2 = os.path.join(self.file_dir[0], 'simple2_test.py')
        simpleClass_test = os.path.join(self.file_dir[0], 'simpleClass_test.py')
        simpleModule_test = os.path.join(self.file_dir[0], 'simpleModule_test.py')

        files_to_tests = {}
        files_to_tests.setdefault(simple_test , []).append('SampleTest.test_xxxxxx1')
        files_to_tests.setdefault(simple_test , []).append('SampleTest.test_xxxxxx2')
        files_to_tests.setdefault(simple_test , []).append('SampleTest.test_non_unique_name')
        files_to_tests.setdefault(simple_test2, []).append('YetAnotherSampleTest.test_abc')
        files_to_tests.setdefault(simpleClass_test, []).append('SetUpClassTest.test_blank')
        files_to_tests.setdefault(simpleModule_test, []).append('SetUpModuleTest.test_blank')

        self._setup_scenario(None, files_to_tests=files_to_tests)
        self.MyTestRunner.verbosity = 2

        buf = pydevd_io.start_redirect(keep_original_redirection=False)
        try:
            self.MyTestRunner.run_tests()
            self.assertEqual(8, len(notifications))
            if sys.version_info[:2] <= (2, 6):
                # The setUpClass is not supported in Python 2.6 (thus we have no collection error).
                expected = [
                    ('notifyTest', 'fail', '', 'AssertionError: Fail test 2', simple_test, 'SampleTest.test_xxxxxx1'),
                    ('notifyTest', 'ok', '', '', simple_test2, 'YetAnotherSampleTest.test_abc'),
                    ('notifyTest', 'ok', '', '', simpleClass_test, 'SetUpClassTest.test_blank'),
                    ('notifyTest', 'ok', '', '', simpleModule_test, 'SetUpModuleTest.test_blank'),
                    ('notifyTest', 'ok', '', '', simple_test, 'SampleTest.test_xxxxxx2'),
                    ('notifyTest', 'ok', 'non unique name ran', '', simple_test, 'SampleTest.test_non_unique_name'),
                    ('notifyTestRunFinished',),
                    ('notifyTestsCollected', 6)
                ]
            else:
                expected = [
                        ('notifyTestsCollected', 6),
                        ('notifyTest', 'ok', 'non unique name ran', '', simple_test, 'SampleTest.test_non_unique_name'),
                        ('notifyTest', 'fail', '', 'AssertionError: Fail test 2', simple_test, 'SampleTest.test_xxxxxx1'),
                        ('notifyTest', 'ok', '', '', simple_test, 'SampleTest.test_xxxxxx2'),
                        ('notifyTest', 'ok', '', '', simple_test2, 'YetAnotherSampleTest.test_abc'),
                    ]

                if not IS_JYTHON:
                    if 'samples.simpleClass_test' in str(notifications):
                        expected.append(('notifyTest', 'error', '', 'ValueError: This is an INTENTIONAL value error in setUpClass.',
                                simpleClass_test.replace('/', os.path.sep), 'samples.simpleClass_test.SetUpClassTest <setUpClass>'))
                        expected.append(('notifyTest', 'error', '', 'ValueError: This is an INTENTIONAL value error in setUpModule.',
                                    simpleModule_test.replace('/', os.path.sep), 'samples.simpleModule_test <setUpModule>'))
                    else:
                        expected.append(('notifyTest', 'error', '', 'ValueError: This is an INTENTIONAL value error in setUpClass.',
                                simpleClass_test.replace('/', os.path.sep), 'simpleClass_test.SetUpClassTest <setUpClass>'))
                        expected.append(('notifyTest', 'error', '', 'ValueError: This is an INTENTIONAL value error in setUpModule.',
                                    simpleModule_test.replace('/', os.path.sep), 'simpleModule_test <setUpModule>'))
                else:
                    expected.append(('notifyTest', 'ok', '', '', simpleClass_test, 'SetUpClassTest.test_blank'))
                    expected.append(('notifyTest', 'ok', '', '', simpleModule_test, 'SetUpModuleTest.test_blank'))

                expected.append(('notifyTestRunFinished',))

            expected.sort()
            new_notifications = []
            for notification in expected:
                try:
                    if len(notification) == 6:
                        # Some are binary on Py3.
                        new_notifications.append((
                            notification[0],
                            notification[1],
                            notification[2].encode('latin1'),
                            notification[3].encode('latin1'),
                            notification[4],
                            notification[5],
                        ))
                    else:
                        new_notifications.append(notification)
                except:
                    raise
            expected = new_notifications

            notifications.sort()
            if not IS_JYTHON:
                self.assertEqual(
                    expected,
                    notifications
                )
        finally:
            pydevd_io.end_redirect()
        b = buf.getvalue()
        if sys.version_info[:2] > (2, 6):
            self.assertTrue(b.find('Ran 4 tests in ') != -1, 'Found: ' + b)
        else:
            self.assertTrue(b.find('Ran 6 tests in ') != -1, 'Found: ' + b)



def complete(completion_context):
    """
    :param CompletionContext completion_context:
    """
    from robocorp_ls_core.lsp import CompletionItemKind
    from robocorp_ls_core.lsp import CompletionItem
    from robocorp_ls_core.lsp import TextEdit
    from robocorp_ls_core.lsp import Range
    from robocorp_ls_core.lsp import Position
    from robotframework_ls.impl import text_utilities
    from robotframework_ls.impl.string_matcher import RobotStringMatcher
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
    )
    from robotframework_ls.impl.robot_lsp_constants import (
        OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL,
    )

    selection = completion_context.sel  #: :type selection: DocumentSelection
    line_start = selection.line_to_column
    items = []

    if line_start:
        tu = text_utilities.TextUtilities(line_start)

        if tu.strip_leading_chars("*"):  # i.e.: the line must start with '*'
            tu.strip()

            words = completion_context.get_accepted_section_header_words()
            config = completion_context.config

            form = config.get_setting(
                OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM,
                str,
                OPTION_ROBOT_COMPLETION_SECTION_HEADERS_FORM_PLURAL,
            )
            matcher = RobotStringMatcher(tu.text)
            for word in words:
                if form == "plural":
                    if not word.endswith("s"):
                        continue
                elif form == "singular":
                    if word.endswith("s"):
                        continue
                if matcher.accepts(word):
                    label = "*** %s ***" % (word,)
                    text_edit = TextEdit(
                        Range(
                            # i.e.: always replace from the start of the line.
                            start=Position(selection.line, 0),
                            end=Position(selection.line, selection.col),
                        ),
                        label,
                    )
                    # text_edit = None
                    items.append(
                        CompletionItem(
                            label,
                            kind=CompletionItemKind.Class,
                            text_edit=text_edit,
                            insertText=text_edit.newText,
                        )
                    )

    return [item.to_dict() for item in items]



"""
Methods for transforming Robot Framework ast model programmatically.
"""
from typing import Optional

from robotidy.app import Robotidy
from robotidy.cli import find_and_read_config, TransformType, validate_regex
from robotidy.files import DEFAULT_EXCLUDES
from robotidy.utils import GlobalFormattingConfig


class RobotidyAPI(Robotidy):
    def __init__(self, src: str, output: Optional[str], **kwargs):
        config = find_and_read_config((src,))
        config = {k: str(v) if not isinstance(v, (list, dict)) else v for k, v in config.items()}
        converter = TransformType()
        transformers = [converter.convert(tr, None, None) for tr in config.get("transform", ())]
        configurations = [converter.convert(c, None, None) for c in config.get("configure", ())]
        formatting_config = GlobalFormattingConfig(
            space_count=kwargs.get("spacecount", None) or int(config.get("spacecount", 4)),
            separator=kwargs.get("separator", None) or config.get("separator", "space"),
            line_sep=config.get("lineseparator", "native"),
            start_line=kwargs.get("startline", None) or int(config["startline"]) if "startline" in config else None,
            end_line=kwargs.get("endline", None) or int(config["endline"]) if "endline" in config else None,
            line_length=kwargs.get("line_length", None) or int(config.get("line_length", 120)),
        )
        exclude = config.get("exclude", None)
        extend_exclude = config.get("extend_exclude", None)
        exclude = validate_regex(exclude if exclude is not None else DEFAULT_EXCLUDES)
        extend_exclude = validate_regex(extend_exclude)
        super().__init__(
            transformers=transformers,
            transformers_config=configurations,
            src=(),
            exclude=exclude,
            extend_exclude=extend_exclude,
            overwrite=False,
            show_diff=False,
            formatting_config=formatting_config,
            verbose=False,
            check=False,
            output=output,
            force_order=False,
        )


def transform_model(model, root_dir: str, output: Optional[str] = None, **kwargs) -> Optional[str]:
    """
    :param model: The model to be transformed.
    :param root_dir: Root directory. Configuration file is searched based
    on this directory or one of its parents.
    :param output: Path where transformed model should be saved
    :param kwargs: Default values for global formatting parameters
    such as ``spacecount``, ``startline`` and ``endline``.
    :return: The transformed model converted to string or None if no transformation took place.
    """
    transformer = RobotidyAPI(root_dir, output, **kwargs)
    diff, _, new_model = transformer.transform(model)
    if not diff:
        return None
    return new_model.text



'''
After breaking on the thread 1, thread 2 should pause waiting for the event1 to be set,
so, when we step return on thread 1, the program should finish if all threads are resumed
or should keep waiting for the thread 2 to run if only thread 1 is resumed.
'''

import threading

event0 = threading.Event()
event1 = threading.Event()
event2 = threading.Event()
event3 = threading.Event()


def _thread1():
    _event1_set = False
    _event2_set = False

    while not event0.is_set():
        event0.wait(timeout=.001)

    event1.set()  # Break thread 1
    _event1_set = True

    while not event2.is_set():
        event2.wait(timeout=.001)
    _event2_set = True  # Note: we can only get here if thread 2 is also released.

    event3.set()


def _thread2():
    event0.set()

    while not event1.is_set():
        event1.wait(timeout=.001)

    event2.set()

    while not event3.is_set():
        event3.wait(timeout=.001)


if __name__ == '__main__':
    threads = [
        threading.Thread(target=_thread1, name='thread1'),
        threading.Thread(target=_thread2, name='thread2'),
    ]
    for t in threads:
        t.start()

    for t in threads:
        t.join()

    print('TEST SUCEEDED!')



"""
Click is a simple Python module inspired by the stdlib optparse to make
writing command line scripts fun. Unlike other modules, it's based
around a simple API that does not come with too much magic and is
composable.
"""
from .core import Argument as Argument
from .core import BaseCommand as BaseCommand
from .core import Command as Command
from .core import CommandCollection as CommandCollection
from .core import Context as Context
from .core import Group as Group
from .core import MultiCommand as MultiCommand
from .core import Option as Option
from .core import Parameter as Parameter
from .decorators import argument as argument
from .decorators import command as command
from .decorators import confirmation_option as confirmation_option
from .decorators import group as group
from .decorators import help_option as help_option
from .decorators import make_pass_decorator as make_pass_decorator
from .decorators import option as option
from .decorators import pass_context as pass_context
from .decorators import pass_obj as pass_obj
from .decorators import password_option as password_option
from .decorators import version_option as version_option
from .exceptions import Abort as Abort
from .exceptions import BadArgumentUsage as BadArgumentUsage
from .exceptions import BadOptionUsage as BadOptionUsage
from .exceptions import BadParameter as BadParameter
from .exceptions import ClickException as ClickException
from .exceptions import FileError as FileError
from .exceptions import MissingParameter as MissingParameter
from .exceptions import NoSuchOption as NoSuchOption
from .exceptions import UsageError as UsageError
from .formatting import HelpFormatter as HelpFormatter
from .formatting import wrap_text as wrap_text
from .globals import get_current_context as get_current_context
from .parser import OptionParser as OptionParser
from .termui import clear as clear
from .termui import confirm as confirm
from .termui import echo_via_pager as echo_via_pager
from .termui import edit as edit
from .termui import getchar as getchar
from .termui import launch as launch
from .termui import pause as pause
from .termui import progressbar as progressbar
from .termui import prompt as prompt
from .termui import secho as secho
from .termui import style as style
from .termui import unstyle as unstyle
from .types import BOOL as BOOL
from .types import Choice as Choice
from .types import DateTime as DateTime
from .types import File as File
from .types import FLOAT as FLOAT
from .types import FloatRange as FloatRange
from .types import INT as INT
from .types import IntRange as IntRange
from .types import ParamType as ParamType
from .types import Path as Path
from .types import STRING as STRING
from .types import Tuple as Tuple
from .types import UNPROCESSED as UNPROCESSED
from .types import UUID as UUID
from .utils import echo as echo
from .utils import format_filename as format_filename
from .utils import get_app_dir as get_app_dir
from .utils import get_binary_stream as get_binary_stream
from .utils import get_text_stream as get_text_stream
from .utils import open_file as open_file

__version__ = "8.1.0"



"""
Duplications checkers
"""
from collections import defaultdict

from robot.api import Token

from robocop.checkers import VisitorChecker
from robocop.rules import Rule, RuleParam, RuleSeverity
from robocop.utils import ROBOT_VERSION, normalize_robot_name, normalize_robot_var_name, get_errors


def configure_sections_order(value):
    section_map = {
        "settings": Token.SETTING_HEADER,
        "variables": Token.VARIABLE_HEADER,
        "testcase": Token.TESTCASE_HEADER,
        "testcases": Token.TESTCASE_HEADER,
        "task": "TASK HEADER",
        "tasks": "TASK HEADER",
        "keyword": Token.KEYWORD_HEADER,
        "keywords": Token.KEYWORD_HEADER,
    }
    sections_order = {}
    for index, name in enumerate(value.split(",")):
        if name.lower() not in section_map or section_map[name.lower()] in sections_order:
            raise ValueError(f"Invalid section name: `{name}`")
        sections_order[section_map[name.lower()]] = index
    if Token.TESTCASE_HEADER in sections_order:
        sections_order["TASK HEADER"] = sections_order[Token.TESTCASE_HEADER]
    return sections_order


rules = {
    "0801": Rule(
        rule_id="0801",
        name="duplicated-test-case",
        msg="Multiple test cases with name '{{ name }}' (first occurrence in line {{ first_occurrence_line }})",
        severity=RuleSeverity.ERROR,
        docs="""
        It is not allowed to reuse the same name of the test case within the same suite in Robot Framework. 
        Name matching is case-insensitive and ignores spaces and underscore characters.
        Duplicated test cases example::
        
            *** Test Cases ***
            Test with name
                No Operation
            
            test_with Name  # it is a duplicate of 'Test with name'
                No Operation
        """,
    ),
    "0802": Rule(
        rule_id="0802",
        name="duplicated-keyword",
        msg="Multiple keywords with name '{{ name }}' (first occurrence in line {{ first_occurrence_line }})",
        severity=RuleSeverity.ERROR,
        docs="""
        Do not define keywords with the same name inside the same file. Name matching is case-insensitive and 
        ignores spaces and underscore characters.
        Duplicated keyword names example::
        
            *** Keywords ***
            Keyword
                No Operaton
            
            keyword
                No Operation
            
            K_eywor d
                No Operation
            
        """,
    ),
    "0803": Rule(
        rule_id="0803",
        name="duplicated-variable",
        msg="Multiple variables with name '{{ name }}' in Variables section (first occurrence in line "
        "{{ first_occurrence_line }}). "
        "Note that Robot Framework is case-insensitive",
        severity=RuleSeverity.ERROR,
        docs="""
        Variable names in Robot Framework are case-insensitive and ignore spaces and underscores. Following variables 
        are duplicates::
        
            *** Variables ***
            ${variable}    1
            ${VARIAble}    a
            @{variable}    a  b
            ${v ariabl e}  c
            ${v_ariable}   d

        """,
    ),
    "0804": Rule(
        rule_id="0804",
        name="duplicated-resource",
        msg="Multiple resource imports with path '{{ name }}' (first occurrence in line {{ first_occurrence_line }})",
        severity=RuleSeverity.WARNING,
    ),
    "0805": Rule(
        rule_id="0805",
        name="duplicated-library",
        msg="Multiple library imports with name '{{ name }}' and identical arguments (first occurrence in line "
        "{{ first_occurrence_line }})",
        severity=RuleSeverity.WARNING,
        docs="""
        If you need to reimport library use alias::
        
            *** Settings ***
            Library  RobotLibrary
            Library  RobotLibrary  AS  OtherRobotLibrary

        """,
    ),
    "0806": Rule(
        rule_id="0806",
        name="duplicated-metadata",
        msg="Duplicated metadata '{{ name }}' (first occurrence in line {{ first_occurrence_line }})",
        severity=RuleSeverity.WARNING,
    ),
    "0807": Rule(
        rule_id="0807",
        name="duplicated-variables-import",
        msg="Duplicated variables import with path '{{ name }}' (first occurrence in line {{ first_occurrence_line }})",
        severity=RuleSeverity.WARNING,
    ),
    "0808": Rule(
        rule_id="0808",
        name="section-already-defined",
        msg="'{{ section_name }}' section header already defined in file (first occurrence in line "
        "{{ first_occurrence_line }})",
        severity=RuleSeverity.WARNING,
        docs="""
        Duplicated section in the file. Robot Framework will handle repeated sections but it is recommended to not 
        duplicate them.
        
        Example::
        
            *** Test Cases ***
            My Test
                Keyword
            
            *** Keywords ***
            Keyword
                No Operation
            
            *** Test Cases ***  # duplicate
            Other Test
                Keyword

        """,
    ),
    "0809": Rule(
        RuleParam(
            name="sections_order",
            default="settings,variables,testcases,keywords",
            converter=configure_sections_order,
            desc="order of sections in comma-separated list",
        ),
        rule_id="0809",
        name="section-out-of-order",
        msg="'{{ section_name }}' section header is defined in wrong order: {{ recommended_order }}",
        severity=RuleSeverity.WARNING,
        docs="""
        Sections should be defined in order set by `sections_order` 
        parameter (default: `settings,variables,testcases,keywords`).
        
        To change the default order use following option::
        
            robocop --configure section-out-of-order:sections_order:comma,separated,list,of,sections
        
        where section should be case-insensitive name from the list: comments, settings, variables, testcases, keywords. 
        Order of not configured sections is ignored.
        
        Example::
        
            *** Settings ***
            
            *** Keywords ***
            
            *** Test Cases ***  # it will report issue because Test Cases should be defined before Keywords

        """,
    ),
    "0810": Rule(
        rule_id="0810",
        name="both-tests-and-tasks",
        msg="Both Task(s) and Test Case(s) section headers defined in file",
        severity=RuleSeverity.ERROR,
        docs="""
        The file contains both Test Case and Task sections. Use only one of them. ::
        
            *** Test Cases ***
            
            *** Tasks ***

        """,
    ),
    "0811": Rule(
        rule_id="0811",
        name="duplicated-argument-name",
        msg="Argument name '{{ argument_name }}' is already used",
        severity=RuleSeverity.ERROR,
        docs="""
        Variable names in Robot Framework are case-insensitive and ignores spaces and underscores. Following arguments 
        are duplicates::
        
            *** Keywords ***
            Keyword
                [Arguments]    ${var}  ${VAR}  ${v_ar}  ${v ar}
                Other Keyword

        """,
    ),
    "0812": Rule(
        rule_id="0812",
        name="duplicated-assigned-var-name",
        msg="Assigned variable name '{{ variable_name }}' is already used",
        severity=RuleSeverity.INFO,
        docs="""
        Variable names in Robot Framework are case-insensitive and ignores spaces and underscores. Following variables 
        are duplicates::
        
            *** Test Cases ***
            Test
                ${var}  ${VAR}  ${v_ar}  ${v ar}  Keyword
        
        """,
    ),
    "0813": Rule(
        rule_id="0813",
        name="duplicated-setting",
        msg="{{ error_msg }}",
        severity=RuleSeverity.WARNING,
        docs="""
        Some settings can be used only once in a file. Only the first value is used.
        Example::
        
            *** Settings ***
            Force Tags        F1
            Force Tags        F2  # this setting will be ignored
        
        """,
    ),
}


class DuplicationsChecker(VisitorChecker):
    """Checker for duplicated names."""

    reports = (
        "duplicated-test-case",
        "duplicated-keyword",
        "duplicated-variable",
        "duplicated-resource",
        "duplicated-library",
        "duplicated-metadata",
        "duplicated-variables-import",
        "duplicated-argument-name",
        "duplicated-assigned-var-name",
        "duplicated-setting",
    )

    def __init__(self):
        self.test_cases = defaultdict(list)
        self.keywords = defaultdict(list)
        self.variables = defaultdict(list)
        self.resources = defaultdict(list)
        self.libraries = defaultdict(list)
        self.metadata = defaultdict(list)
        self.variable_imports = defaultdict(list)
        super().__init__()

    def visit_File(self, node):  # noqa
        self.test_cases = defaultdict(list)
        self.keywords = defaultdict(list)
        self.variables = defaultdict(list)
        self.resources = defaultdict(list)
        self.libraries = defaultdict(list)
        self.metadata = defaultdict(list)
        self.variable_imports = defaultdict(list)
        super().visit_File(node)
        self.check_duplicates(self.test_cases, "duplicated-test-case")
        self.check_duplicates(self.keywords, "duplicated-keyword")
        self.check_duplicates(self.variables, "duplicated-variable")
        self.check_duplicates(self.resources, "duplicated-resource")
        self.check_duplicates(self.libraries, "duplicated-library")
        self.check_duplicates(self.metadata, "duplicated-metadata")
        self.check_duplicates(self.variable_imports, "duplicated-variables-import")

    def check_duplicates(self, container, rule):
        for nodes in container.values():
            for duplicate in nodes[1:]:
                self.report(rule, name=duplicate.name, first_occurrence_line=nodes[0].lineno, node=duplicate)

    def visit_TestCase(self, node):  # noqa
        testcase_name = normalize_robot_name(node.name)
        self.test_cases[testcase_name].append(node)
        self.generic_visit(node)

    def visit_Keyword(self, node):  # noqa
        keyword_name = normalize_robot_name(node.name)
        self.keywords[keyword_name].append(node)
        self.generic_visit(node)

    def visit_KeywordCall(self, node):  # noqa
        assign = node.get_tokens(Token.ASSIGN)
        seen = set()
        for var in assign:
            name = normalize_robot_var_name(var.value)
            if name in seen:
                self.report(
                    "duplicated-assigned-var-name",
                    variable_name=var.value,
                    node=node,
                    lineno=var.lineno,
                    col=var.col_offset + 1,
                )
            else:
                seen.add(name)

    def visit_VariableSection(self, node):  # noqa
        self.generic_visit(node)

    def visit_Variable(self, node):  # noqa
        if not node.name or get_errors(node):
            return
        var_name = normalize_robot_name(self.replace_chars(node.name, "${}@&"))
        self.variables[var_name].append(node)

    @staticmethod
    def replace_chars(name, chars):
        return "".join(c for c in name if c not in chars)

    def visit_ResourceImport(self, node):  # noqa
        if node.name:
            self.resources[node.name].append(node)

    def visit_LibraryImport(self, node):  # noqa
        if not node.name:
            return
        name_with_args = node.name + "".join(token.value for token in node.data_tokens[2:])
        self.libraries[name_with_args].append(node)

    def visit_Metadata(self, node):  # noqa
        if node.name is not None:
            self.metadata[node.name + node.value].append(node)

    def visit_VariablesImport(self, node):  # noqa
        if not node.name:
            return
        # only python files can have arguments - covered in E0404 variables-import-with-args
        if not node.name.endswith(".py") and node.get_token(Token.ARGUMENT):
            return
        name_with_args = node.name + "".join(token.value for token in node.data_tokens[2:])
        self.variable_imports[name_with_args].append(node)

    def visit_Arguments(self, node):  # noqa
        args = set()
        for arg in node.get_tokens(Token.ARGUMENT):
            orig, *_ = arg.value.split("=", maxsplit=1)
            name = normalize_robot_var_name(orig)
            if name in args:
                self.report(
                    "duplicated-argument-name",
                    argument_name=orig,
                    node=node,
                    lineno=arg.lineno,
                    col=arg.col_offset + 1,
                )
            else:
                args.add(name)

    def visit_Error(self, node):  # noqa
        for error in get_errors(node):
            self.report("duplicated-setting", error_msg=error, node=node)


class SectionHeadersChecker(VisitorChecker):
    """Checker for duplicated or out of order section headers."""

    reports = (
        "section-already-defined",
        "section-out-of-order",
        "both-tests-and-tasks",
    )

    def __init__(self):
        self.sections_by_order = []
        self.sections_by_existence = dict()
        super().__init__()

    @staticmethod
    def section_order_to_str(order):
        by_index = sorted([(key, value) for key, value in order.items()], key=lambda x: x[1])
        name_map = {
            Token.SETTING_HEADER: "Settings",
            Token.VARIABLE_HEADER: "Variables",
            Token.TESTCASE_HEADER: "Test Cases / Tasks",
            "TASK HEADER": "Test Cases / Tasks",
            Token.KEYWORD_HEADER: "Keywords",
        }
        order_str = []
        for name, _ in by_index:
            mapped_name = name_map[name]
            if mapped_name not in order_str:
                order_str.append(mapped_name)
        return " > ".join(order_str)

    def visit_File(self, node):  # noqa
        self.sections_by_order = []
        self.sections_by_existence = dict()
        super().visit_File(node)

    def visit_SectionHeader(self, node):  # noqa
        section_name = node.type
        if section_name not in self.param("section-out-of-order", "sections_order"):
            return
        if section_name == Token.TESTCASE_HEADER:
            if "task" in node.name.lower():
                section_name = "TASK HEADER"
                if Token.TESTCASE_HEADER in self.sections_by_existence:
                    self.report("both-tests-and-tasks", node=node)
            else:
                if "TASK HEADER" in self.sections_by_existence:
                    self.report("both-tests-and-tasks", node=node)
        order_id = self.param("section-out-of-order", "sections_order")[section_name]
        if section_name in self.sections_by_existence:
            self.report(
                "section-already-defined",
                section_name=node.data_tokens[0].value,
                first_occurrence_line=self.sections_by_existence[section_name],
                node=node,
            )
        else:
            self.sections_by_existence[section_name] = node.lineno
        if any(previous_id > order_id for previous_id in self.sections_by_order):
            self.report(
                "section-out-of-order",
                section_name=node.data_tokens[0].value,
                recommended_order=self.section_order_to_str(self.param("section-out-of-order", "sections_order")),
                node=node,
            )
        self.sections_by_order.append(order_id)



class A(object):

    def __init__(self):
        self.a = 10


class B(A):

    def __init__(self):
        super().__init__()  # break here
        assert self.a == 10

        def method():
            self.b = self.a

        method()
        assert self.b == 10


if __name__ == '__main__':
    B()
    print('TEST SUCEEDED')



from __future__ import print_function
from multiprocessing import Pool
from time import sleep


def call(arg):
    if arg == '2':
        print("called")  # break 1 here


if __name__ == '__main__':
    pool = Pool(1)
    pool.map(call, ['1', '2', '3'])
    pool.close()
    pool.join()
    sleep(1)
    print('TEST SUCEEDED!')  # break 2 here



import sys


def main():
    import subprocess
    process = subprocess.Popen(
        'git status --porcelain'.split(), stderr=subprocess.STDOUT, stdout=subprocess.PIPE)
    output, _ = process.communicate()
    if output:
        if sys.version_info[0] > 2:
            output = output.decode('utf-8')

        files = set()
        for line in output.splitlines():
            filename = line[3:]
            files.add(filename.strip())

        files.discard('.travis_install_python_deps.sh')
        files.discard('miniconda.sh')
        if files:
            # If there are modifications, show a diff of the modifications and fail the script.
            # (we're mostly interested in modifications to the .c generated files by cython).
            print('Found modifications in git:\n%s ' % (output,))
            print('Files: %s' % (files,))
            print('----------- diff -------------')
            subprocess.call('git diff'.split())
            print('----------- end diff -------------')
            sys.exit(1)


if __name__ == '__main__':
    main()



from typing import List

from robocorp_ls_core.protocols import check_implements
from robocorp_ls_core.robotframework_log import get_logger
from robotframework_ls.impl.protocols import (
    ICompletionContext,
    IKeywordFound,
    IKeywordCollector,
    AbstractKeywordCollector,
    KeywordUsageInfo,
)
from robocorp_ls_core.lsp import (
    TextEditTypedDict,
    CompletionItemTypedDict,
    InsertTextFormat,
)


log = get_logger(__name__)


class _Collector(AbstractKeywordCollector):
    def __init__(
        self, completion_context: ICompletionContext, keyword_usage: KeywordUsageInfo
    ):
        from robotframework_ls.impl.string_matcher import RobotStringMatcher
        from robotframework_ls.impl.string_matcher import (
            build_matchers_with_resource_or_library_scope,
        )
        from robotframework_ls.robot_config import create_convert_keyword_format_func

        token_str = keyword_usage.token.value

        self.completion_items: List[CompletionItemTypedDict] = []
        self.completion_context = completion_context
        self.selection = completion_context.sel
        self.token = keyword_usage.token

        # i.e.: if we have arguments already don't add arguments again.
        self._add_arguments = True
        for t in keyword_usage.node.tokens:
            if t.type == t.ARGUMENT and t.value:
                self._add_arguments = False
                break

        self._matcher = RobotStringMatcher(token_str)
        self._scope_matchers = build_matchers_with_resource_or_library_scope(token_str)
        config = completion_context.config

        self._convert_keyword_format = create_convert_keyword_format_func(config)

    def accepts(self, keyword_name: str) -> bool:
        if self._matcher.accepts_keyword_name(keyword_name):
            return True
        for matcher in self._scope_matchers:
            if matcher.accepts_keyword_name(keyword_name):
                return True
        return False

    def _create_completion_item_from_keyword(
        self, keyword_found: IKeywordFound, selection, token, col_delta=0
    ) -> CompletionItemTypedDict:
        from robotframework_ls.impl.protocols import IKeywordArg

        label = keyword_found.keyword_name

        if keyword_found.library_name:
            # If we found the keyword in a library, convert its format depending on
            # the user configuration.
            label = self._convert_keyword_format(label)

        text = label

        if self._add_arguments:
            arg: IKeywordArg
            for i, arg in enumerate(keyword_found.keyword_args):
                if (
                    arg.is_keyword_arg
                    or arg.is_star_arg
                    or arg.default_value is not None
                ):
                    continue

                arg_name = arg.arg_name
                arg_name = (
                    arg_name.replace("$", "\\$").replace("{", "").replace("}", "")
                )

                text += "    ${%s:%s}" % (i + 1, arg_name)

        text_edit: TextEditTypedDict = {
            "range": {
                "start": {
                    "line": selection.line,
                    "character": token.col_offset + col_delta,
                },
                "end": {"line": selection.line, "character": token.end_col_offset},
            },
            "newText": text,
        }

        if keyword_found.library_name:
            label = f"{label} ({keyword_found.library_name})"

        elif keyword_found.resource_name:
            label = f"{label} ({keyword_found.resource_name})"

        return {
            "label": label,
            "kind": keyword_found.completion_item_kind,
            "textEdit": text_edit,
            "insertText": text_edit["newText"],
            "insertTextFormat": InsertTextFormat.Snippet,
        }

    def on_keyword(self, keyword_found):
        col_delta = 0
        if not self._matcher.accepts_keyword_name(keyword_found.keyword_name):
            for matcher in self._scope_matchers:
                if matcher.accepts_keyword(keyword_found):
                    # +1 for the dot
                    col_delta = len(matcher.resource_or_library_name) + 1
                    break
            else:
                return  # i.e.: don't add completion

        item = self._create_completion_item_from_keyword(
            keyword_found, self.selection, self.token, col_delta=col_delta
        )

        self.completion_context.assign_documentation_resolve(
            item, keyword_found.compute_docs_with_signature
        )

        self.completion_items.append(item)

    def __typecheckself__(self) -> None:
        _: IKeywordCollector = check_implements(self)


def complete(completion_context: ICompletionContext) -> List[CompletionItemTypedDict]:
    from robotframework_ls.impl.collect_keywords import collect_keywords
    from robotframework_ls.impl import ast_utils

    token_info = completion_context.get_current_token()
    if token_info is not None:
        keyword_usage = ast_utils.create_keyword_usage_info_from_token(
            token_info.stack, token_info.node, token_info.token
        )
        if keyword_usage is not None and not keyword_usage.token.value.strip().endswith(
            "}"
        ):
            collector = _Collector(completion_context, keyword_usage)
            collect_keywords(completion_context, collector)

            return collector.completion_items

    return []





def check():
    import pydevd
    py_db = pydevd.get_global_debugger()
    assert len(py_db.api_received_breakpoints) == 1
    _a = 10  # break here
    # should remove the breakpoints when stopped on the previous line
    assert len(py_db.api_received_breakpoints) == 0


if __name__ == '__main__':
    check()
    print('TEST SUCEEDED!')



#!/usr/bin/env python
"""Simple GLUT example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for glut
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console
4) run: gl.glClearColor(1,1,1,1)
"""

if __name__ == '__main__':

    #!/usr/bin/env python
    import sys
    import OpenGL.GL as gl
    import OpenGL.GLUT as glut
    
    def close():
        glut.glutDestroyWindow(glut.glutGetWindow())
    
    def display():
        gl.glClear (gl.GL_COLOR_BUFFER_BIT | gl.GL_DEPTH_BUFFER_BIT)
        glut.glutSwapBuffers()
    
    def resize(width,height):
        gl.glViewport(0, 0, width, height+4)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadIdentity()
        gl.glOrtho(0, width, 0, height+4, -1, 1)
        gl.glMatrixMode(gl.GL_MODELVIEW)
    
    if glut.glutGetWindow() > 0:
        interactive = True
        glut.glutInit(sys.argv)
        glut.glutInitDisplayMode(glut.GLUT_DOUBLE |
                                 glut.GLUT_RGBA   |
                                 glut.GLUT_DEPTH)
    else:
        interactive = False
    
    glut.glutCreateWindow('gui-glut')
    glut.glutDisplayFunc(display)
    glut.glutReshapeFunc(resize)
    # This is necessary on osx to be able to close the window
    #  (else the close button is disabled)
    if sys.platform == 'darwin' and not bool(glut.HAVE_FREEGLUT):
        glut.glutWMCloseFunc(close)
    gl.glClearColor(0,0,0,1)
    
    if not interactive:
        glut.glutMainLoop()



import unittest

class Test(unittest.TestCase):

    def testProcessCommandLine(self):
        from _pydevd_bundle.pydevd_command_line_handling import process_command_line, setup_to_argv
        setup = process_command_line(['pydevd.py', '--port', '1', '--save-threading'])
        assert setup['save-threading']
        assert setup['port'] == 1
        assert not setup['qt-support']

        argv = setup_to_argv(setup)
        assert argv[0].endswith('pydevd.py') or argv[0].endswith('pydevd$py.class'), 'Expected: %s to end with pydevd.py' % (argv[0],)
        argv = argv[1:]
        assert argv == ['--port', '1', '--save-threading']

    def testProcessCommandLine2(self):
        from _pydevd_bundle.pydevd_command_line_handling import process_command_line, setup_to_argv
        setup = process_command_line(['pydevd.py', '--port', '1', '--qt-support=auto'])
        assert setup['qt-support'] == 'auto'

        setup = process_command_line(['pydevd.py', '--port', '1', '--qt-support'])
        assert setup['qt-support'] == 'auto'

        setup = process_command_line(['pydevd.py', '--port', '1', '--qt-support=pyqt4'])
        assert setup['qt-support'] == 'pyqt4'

        self.assertRaises(ValueError, process_command_line, ['pydevd.py', '--port', '1', '--qt-support=wrong'])



#!/usr/bin/env python
"""Simple Qt4 example to manually test event loop integration.

To run this:
1) Enable the PyDev GUI event loop integration for qt
2) do an execfile on this script
3) ensure you have a working GUI simultaneously with an
   interactive console

Ref: Modified from http://zetcode.com/tutorials/pyqt4/firstprograms/
"""

if __name__ == '__main__':
    import sys
    from PyQt4 import QtGui, QtCore
    
    class SimpleWindow(QtGui.QWidget):
        def __init__(self, parent=None):
            QtGui.QWidget.__init__(self, parent)
    
            self.setGeometry(300, 300, 200, 80)
            self.setWindowTitle('Hello World')
    
            quit = QtGui.QPushButton('Close', self)
            quit.setGeometry(10, 10, 60, 35)
    
            self.connect(quit, QtCore.SIGNAL('clicked()'),
                         self, QtCore.SLOT('close()'))
    
    if __name__ == '__main__':
        app = QtCore.QCoreApplication.instance()
        if app is None:
            app = QtGui.QApplication([])
    
        sw = SimpleWindow()
        sw.show()



from _pydevd_bundle._debug_adapter.pydevd_schema import InitializeRequest, \
    InitializeRequestArguments, InitializeResponse, Capabilities
from _pydevd_bundle._debug_adapter import pydevd_schema, pydevd_base_schema
from _pydevd_bundle._debug_adapter.pydevd_schema import ThreadsResponse


def test_schema():

    json_msg = '''
{
    "arguments": {
        "adapterID": "pydevd",
        "clientID": "vscode",
        "clientName": "Visual Studio Code",
        "columnsStartAt1": true,
        "linesStartAt1": true,
        "locale": "en-us",
        "pathFormat": "path",
        "supportsRunInTerminalRequest": true,
        "supportsVariablePaging": true,
        "supportsVariableType": true
    },
    "command": "initialize",
    "seq": 1,
    "type": "request"
}'''

    initialize_request = pydevd_base_schema.from_json(json_msg)
    assert initialize_request.__class__ == InitializeRequest
    assert initialize_request.arguments.__class__ == InitializeRequestArguments
    assert initialize_request.arguments.adapterID == 'pydevd'
    assert initialize_request.command == 'initialize'
    assert initialize_request.type == 'request'
    assert initialize_request.seq == 1

    response = pydevd_base_schema.build_response(initialize_request)
    assert response.__class__ == InitializeResponse
    assert response.seq == -1  # Must be set before sending
    assert response.command == 'initialize'
    assert response.type == 'response'
    assert response.body.__class__ == Capabilities

    assert response.to_dict() == {
        "seq":-1,
        "type": "response",
        "request_seq": 1,
        "success": True,
        "command": "initialize",
        "body": {}
    }

    capabilities = response.body  # : :type capabilities: Capabilities
    capabilities.supportsCompletionsRequest = True
    assert response.to_dict() == {
        "seq":-1,
        "type": "response",
        "request_seq": 1,
        "success": True,
        "command": "initialize",
        "body": {'supportsCompletionsRequest':True}
    }

    initialize_event = pydevd_schema.InitializedEvent()
    assert initialize_event.to_dict() == {
        "seq":-1,
        "type": "event",
        "event": "initialized"
    }


def test_schema_translation_frame():
    pydevd_base_schema.BaseSchema.initialize_ids_translation()
    stack_trace_arguments = pydevd_schema.StackTraceArguments(threadId=1)
    stack_trace_request = pydevd_schema.StackTraceRequest(stack_trace_arguments)

    stackFrames = [
        pydevd_schema.StackFrame(id=2 ** 45, name='foo', line=1, column=1).to_dict(),
        pydevd_schema.StackFrame(id=2 ** 46, name='bar', line=1, column=1).to_dict(),
    ]
    body = pydevd_schema.StackTraceResponseBody(stackFrames)
    stack_trace_response = pydevd_base_schema.build_response(stack_trace_request, kwargs=dict(body=body))
    as_dict = stack_trace_response.to_dict(update_ids_to_dap=True)
    assert as_dict == {
        'type': 'response',
        'request_seq':-1,
        'success': True,
        'command': 'stackTrace',
        'body': {'stackFrames': [
            {'id': 1, 'name': 'foo', 'line': 1, 'column': 1, 'source': {}},
            {'id': 2, 'name': 'bar', 'line': 1, 'column': 1, 'source': {}},
        ]},
        'seq':-1}

    reconstructed = pydevd_base_schema.from_dict(as_dict, update_ids_from_dap=True)
    assert reconstructed.to_dict() == {
        'type': 'response',
        'request_seq':-1,
        'success': True,
        'command': 'stackTrace',
        'body': {'stackFrames': [
            {'id': 2 ** 45, 'name': 'foo', 'line': 1, 'column': 1, 'source': {}},
            {'id': 2 ** 46, 'name': 'bar', 'line': 1, 'column': 1, 'source': {}}
        ]},
        'seq':-1
    }


def test_schema_translation_thread():
    from _pydevd_bundle._debug_adapter.pydevd_schema import ThreadsRequest
    pydevd_base_schema.BaseSchema.initialize_ids_translation()

    threads = [
        pydevd_schema.Thread(id=2 ** 45, name='foo').to_dict(),
        pydevd_schema.Thread(id=2 ** 46, name='bar').to_dict(),
    ]
    body = pydevd_schema.ThreadsResponseBody(threads)
    threads_request = ThreadsRequest()
    threads_response = pydevd_base_schema.build_response(threads_request, kwargs=dict(body=body))
    as_dict = threads_response.to_dict(update_ids_to_dap=True)
    assert as_dict == {
        'type': 'response',
        'request_seq':-1,
        'success': True,
        'command': 'threads',
        'body': {'threads': [
            {'id': 1, 'name': 'foo'},
            {'id': 2, 'name': 'bar'},
        ]},
        'seq':-1}

    reconstructed = pydevd_base_schema.from_dict(as_dict, update_ids_from_dap=True)
    assert reconstructed.to_dict() == {
        'type': 'response',
        'request_seq':-1,
        'success': True,
        'command': 'threads',
        'body': {'threads': [
            {'id': 2 ** 45, 'name': 'foo'},
            {'id': 2 ** 46, 'name': 'bar'}
        ]},
        'seq':-1
    }
